# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Seed inputs for synthetic I/O example generation."""

import enum
import random
from typing import List, Optional

from absl import logging
import numpy as np
import tensorflow as tf
from tf_coder import tf_coder_utils


def _ints_from_nonnegative(ints: List[int]) -> List[int]:
  return [-x for x in ints][:0:-1] + ints


def _np_array_to_float_list(np_array) -> List[float]:
  return [round(float(f), 5) for f in np_array]

# Each element is a list of bools to choose from. For example, when populating a
# boolean tensor, all of the elements should come from the same list of choices.
BOOL_CHOICES = [
    # Different ratios of True and False.
    [False, True],
    [False] * 4 + [True],
    [False] * 9 + [True],
    [True] * 4 + [False],
    [True] * 9 + [False],
]

# Each element is a list of nonnegative ints to choose from.
UINT_CHOICES = [
    # 0s and 1s with different ratios.
    [0, 1],
    [0] * 4 + [1],
    [0] + [1] * 4,

    # Nonnegative ints. (Positive ints can be obtained by "getting lucky" and
    # not choosing 0, which should happen often.)
    list(range(6)),
    list(range(10)),
    list(range(20)),
    list(range(50)),
    list(range(100)),
]

# Each element is a list of ints to choose from.
INT_CHOICES = UINT_CHOICES + [
    # 0s, 1s, -1s.
    [-1, 0, 1],

    # Negative and nonnegative ints.
    list(range(-5, 6)),
    list(range(-9, 10)),
    list(range(-19, 20)),
    list(range(-49, 50)),
    list(range(-99, 100)),
    list(range(-1000, 1001)),
]

# Each element is a list of floats to choose from.
FLOAT_CHOICES = [
    # 0s and 1s with different ratios.
    [0.0, 1.0],
    [0.0] * 4 + [1.0],
    [0.0] + [1.0] * 4,

    # Ints as floats.
    [float(i) for i in range(10)],
    [float(i) for i in range(50)],
    [float(i) for i in range(-9, 10)],
    [float(i) for i in range(-99, 100)],

    # Probabilities with higher chance of exactly 0.0 or 1.0.
    [0.0] * 2 + _np_array_to_float_list(np.arange(0, 1.1, 0.1)) + [1.0] * 2,
    [0.0] * 50 + _np_array_to_float_list(np.arange(0, 1.01, 0.01)) + [1.0] * 50,

    # Nonnegative floats.
    _np_array_to_float_list(np.arange(0, 5, 0.1)),
    _np_array_to_float_list(np.arange(0, 10, 0.1)),
    _np_array_to_float_list(np.arange(0, 50, 0.1)),

    # Negative and nonnegative floats.
    _np_array_to_float_list(np.arange(-4.9, 5, 0.1)),
    _np_array_to_float_list(np.arange(-19.9, 20, 0.1)),
    _np_array_to_float_list(np.arange(-99, 100, 0.1)),
]


def generate_random_tensor(python_type, shape: Optional[List[int]] = None):
  """Returns a random tensor, optionally with the given shape.

  First choose the rank and shape. Depending on the python_type, choose a list
  of value choices from the lists above, decide whether to sample with or
  without replacement from that list of choices, and then randomly construct a
  tensor of the chosen shape using randomly selected values.

  Args:
    python_type: One of `int`, `float`, or `bool`.
    shape: Optional list of ints describing the desired shape, or None to have
      the shape be randomly chosen.
  """
  if shape is None:
    rank = np.random.choice([0, 1, 2, 3, 4], p=[0.1, 0.3, 0.3, 0.2, 0.1])
    # Randomly fill the shape, but make sure there aren't too many elements.
    while True:
      shape = []
      for _ in range(rank):
        length_distribution = 1 / (1 + np.arange(1, 21))
        length_distribution /= np.sum(length_distribution)
        shape.append(np.random.choice(range(1, 21), p=length_distribution))
      if np.prod(shape) <= 100:
        # We're assuming that no one will create a larger input tensor.
        break

  if python_type == bool:
    dtype = tf.bool
    choices_list = BOOL_CHOICES
    sort = False
  elif python_type == float:
    dtype = (
        tf.float32 if random.random() < 0.9 else random.choice(
            tf_coder_utils.FLOAT_DTYPES))
    choices_list = FLOAT_CHOICES
    sort = random.random() < 0.1
  elif python_type == int:
    dtype = (
        tf.int32
        if random.random() < 0.9 else random.choice(tf_coder_utils.INT_DTYPES))
    choices_list = (
        UINT_CHOICES if dtype in tf_coder_utils.UINT_DTYPES else INT_CHOICES)
    sort = random.random() < 0.1
  else:
    logging.fatal('Unhandled python_type: %s', python_type)

  choices = random.choice(choices_list)
  num_elements = np.prod(shape)
  replace = (True if num_elements > len(choices) else
             np.random.choice([False, True], p=[0.3, 0.7]))
  elements = np.random.choice(choices, size=shape, replace=replace)
  tensor = tf.constant(elements, dtype=dtype)

  # Only nonscalar tensors of certain dtypes can be sorted.
  if sort and len(shape) and (python_type == float
                              or dtype in [tf.int32, tf.int64]):
    tensor = tf.sort(tensor)
  return tensor


@enum.unique
class ValueKind(enum.Enum):
  PRIMITIVE = 1
  SEQUENCE = 2
  TENSOR = 3
  SPARSE_TENSOR = 4


def generate_random_input(random_seed: Optional[int] = None):
  """Create a random input."""
  if random_seed is not None:
    random.seed(random_seed)
    np.random.seed(random_seed)

  python_type = np.random.choice([int, float, bool], p=[0.5, 0.4, 0.1])
  kind = np.random.choice([ValueKind.PRIMITIVE, ValueKind.SEQUENCE,
                           ValueKind.TENSOR, ValueKind.SPARSE_TENSOR],
                          p=[0.05, 0.05, 0.8, 0.1])

  if kind == ValueKind.PRIMITIVE:
    scalar_tensor = generate_random_tensor(python_type=python_type, shape=[])
    return python_type(scalar_tensor.numpy())

  elif kind == ValueKind.SEQUENCE:
    # Generate a random tensor and turn it into a sequence along the first
    # dimension.
    tensor = generate_random_tensor(python_type=python_type)
    # An empty TensorShape, when converted to a bool, is True.
    while not len(tensor.shape):  # pylint: disable=g-explicit-length-test
      tensor = generate_random_tensor(python_type=python_type)
    return list(tensor)

  elif kind == ValueKind.TENSOR:
    return generate_random_tensor(python_type=python_type)

  elif kind == ValueKind.SPARSE_TENSOR:
    # 3 and 4 are the most likely, and the probability decreases as the number
    # gets farther from 3.5.
    num_elements_distribution = 1 / (1 + np.abs(np.arange(1, 21) - 3.5))
    num_elements_distribution /= np.sum(num_elements_distribution)
    num_elements = np.random.choice(range(1, 21), p=num_elements_distribution)
    values = generate_random_tensor(python_type=python_type,
                                    shape=[num_elements])

    rank = np.random.choice([1, 2, 3, 4], p=[0.3, 0.3, 0.3, 0.1])
    while True:
      dense_shape = []
      for _ in range(rank):
        dense_shape.append(random.choice(
            list(range(1, 4)) * 5 +
            list(range(1, 10)) +
            list(range(10, 100, 10)) +
            list(range(100, 1001, 100))))
      if np.prod(dense_shape) >= 2 * num_elements:
        break

    indices_set = set()
    while len(indices_set) < num_elements:
      coordinates = tuple(random.choice(range(dense_shape[dimension]))
                          for dimension in range(rank))
      indices_set.add(coordinates)
    indices = list(indices_set)

    # TODO(b/140891723): tf.sparse.reorder doesn't work when the SparseTensor
    # has values of dtype tf.uint32 or tf.uint64. This workaround instead casts
    # the values to a workable dtype.
    if values.dtype in (tf.uint32, tf.uint64):
      values = tf.cast(values, tf.int32)

    sparse_tensor = tf.SparseTensor(values=values, indices=indices,
                                    dense_shape=dense_shape)
    return tf.sparse.reorder(sparse_tensor)

  else:
    logging.fatal('Unhandled kind: %s', kind)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Loads extracted tf_function data from disk."""

import os
from typing import Dict, List, Text

import tensorflow as tf
from tf_coder.datasets.github import tokenizer

ADD_OPERATION_NAMES = True
ADD_OPERATION_DOCSTRINGS = False

DEFAULT_DATA_PREFIX = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   'data', '')


def parse_example_proto(serialized_example: Text) -> Dict[Text, tf.Tensor]:
  """Parses a serialized tensorflow Example into its component tensors.

  Executed in TensorFlow graph mode by tf.data.Dataset.map.

  Args:
    serialized_example: A single tf.Example, serialized as bytes. The output
      of the extract_tf_functions Beam pipeline.
  Returns:
    A dict mapping keys to string tensors for a single example.
  """
  features = {
      'docstring': tf.io.VarLenFeature(tf.string),
      'names': tf.io.VarLenFeature(tf.string),
      'comments': tf.io.VarLenFeature(tf.string),
      'strings': tf.io.VarLenFeature(tf.string),
      'tf_functions': tf.io.VarLenFeature(tf.string),
  }
  parsed = tf.io.parse_single_example(serialized_example, features)
  for key in parsed:
    parsed[key] = tf.sparse.to_dense(parsed[key])
  return parsed


def _as_text_list(value: tf.Tensor) -> List[Text]:
  return [b.decode('utf-8') for b in value.numpy().tolist()]


def _as_python_example(
    example: Dict[Text, tf.Tensor]) -> Dict[Text, List[Text]]:
  return {
      key: _as_text_list(value)
      for key, value in example.items()
  }


def load_data(prefix) -> List[Dict[Text, List[Text]]]:
  filenames = tf.io.gfile.glob(prefix + '*')
  dataset = tf.data.TFRecordDataset(filenames)
  dataset = dataset.map(parse_example_proto)
  return [_as_python_example(example) for example in dataset]


def get_operations(example: Dict[Text, List[Text]]) -> Text:
  return ' '.join(
      [tf_function[3:] if tf_function.startswith('tf.') else tf_function
       for tf_function in example['tf_functions']]
  )


def get_context(example: Dict[Text, List[Text]]) -> Text:
  """Gets the textual context provided in a single example."""
  docstring = example['docstring'][0]
  comments = example['comments']
  names = example['names']
  strings = example['strings']

  tokens = (
      tokenizer.tokenize(docstring)
      + tokenizer.tokens_from_text_list(comments)
      + tokenizer.tokens_from_text_list(names)
      + tokenizer.tokens_from_text_list(strings)
  )
  return ' '.join(tokens)


def get_full_context(example):
  context = get_context(example)
  if ADD_OPERATION_NAMES:
    context += ' ' + get_operations(example)
  if ADD_OPERATION_DOCSTRINGS:
    raise NotImplementedError()
  return ' '.join(tokenizer.tokenize(context))


def uses_operation(example: Dict[Text, List[Text]], tf_function: Text) -> bool:
  return tf_function in example['tf_functions']



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Value search endpoint for the Colab interface."""

from typing import Any, Dict, List, Optional, Text

from tf_coder.benchmarks import all_benchmarks
from tf_coder.models import tensor_features_model
from tf_coder.natural_language import description_handler_factory
from tf_coder.value_search import all_operations
from tf_coder.value_search import value_search
from tf_coder.value_search import value_search_settings as settings_module

# A message to print when things go wrong.
CONTACT_MESSAGE = (
    'If you see this message, please raise a GitHub issue '
    '(https://github.com/google-research/tensorflow-coder/issues) describing '
    'what happened.')

# The global description handler for this session. This will be loaded during
# warm_up().
DESCRIPTION_HANDLER = None

# The tensor features model and config. These will be loaded during warm_up().
TENSOR_MODEL = None
TENSOR_CONFIG = None

# Whether warm_up() was called before.
WARMED_UP = False

# Default settings, if the user doesn't provide their own settings.
DEFAULT_SETTINGS = settings_module.default_settings()


def warm_up():
  """Loads and warms up the tensor features model."""
  global DESCRIPTION_HANDLER, TENSOR_MODEL, TENSOR_CONFIG, WARMED_UP
  if WARMED_UP:
    return
  WARMED_UP = True

  # Use the default choices for the description handler and tensor model. The
  # Colab interface will not allow users to change these.

  DESCRIPTION_HANDLER = description_handler_factory.create_handler(
      DEFAULT_SETTINGS.description_handler_name)
  if (not DEFAULT_SETTINGS.tensor_model.config_path or
      not DEFAULT_SETTINGS.tensor_model.checkpoint_path):
    return
  try:
    TENSOR_CONFIG = tensor_features_model.load_config(
        DEFAULT_SETTINGS.tensor_model.config_path)
    TENSOR_MODEL = tensor_features_model.get_model(TENSOR_CONFIG)
    tensor_checkpoint = tensor_features_model.create_checkpoint(TENSOR_MODEL)
    tensor_checkpoint.restore(
        DEFAULT_SETTINGS.tensor_model.checkpoint_path).expect_partial()

    # Warm up. Running the model for the first time takes an extra ~10 seconds.
    value_search.operation_multipliers_from_tensor_model(
        all_benchmarks.find_benchmark_with_name('simple_cast'),
        TENSOR_MODEL, TENSOR_CONFIG, DEFAULT_SETTINGS)
  except Exception:  # pylint: disable=broad-except
    # No matter what goes wrong with loading the tensor features model, we
    # should fall back to value search without the model.
    print('Could not load the tensor features model. ' + CONTACT_MESSAGE)
    TENSOR_MODEL = None
    TENSOR_CONFIG = None


def run_value_search_from_colab(
    inputs: Dict[Text, Any],
    output: Any,
    constants: Optional[List[Any]] = None,
    description: Optional[Text] = None,
    settings: Optional[settings_module.Settings] = None
) -> value_search.ValueSearchResults:
  """Value search endpoint for the Colab interface.

  Args:
    inputs: A dict mapping input variable names to input tensors.
    output: The corresponding desired output.
    constants: An optional list of scalar constants.
    description: An optional natural language description of the task.
    settings: A Settings object containing settings for the search.

  Returns:
    A ValueSearchResults namedtuple.
  """
  if not WARMED_UP:
    warm_up()

  return value_search.run_value_search_from_example(
      inputs=inputs,
      output=output,
      settings=settings if settings is not None else DEFAULT_SETTINGS,
      constants=constants,
      description=description,
      source='From TF-Coder Colab',
      description_handler=DESCRIPTION_HANDLER,
      tensor_model=TENSOR_MODEL,
      tensor_config=TENSOR_CONFIG)


def print_supported_operations():
  """Prints all of the supported operations."""
  print('TensorFlow functions:\n'
        '---------------------')
  for operation in all_operations.get_tf_operations():
    print(operation.name)
  print()
  print('SparseTensor functions:\n'
        '-----------------------')
  for operation in all_operations.get_sparse_operations():
    print(operation.name)
  print()
  print('Python-syntax operations:\n'
        '-------------------------')
  for operation in all_operations.get_python_operations():
    syntax_form = operation.reconstruct_expression_from_strings(
        ['arg{}'.format(i+1) for i in range(operation.num_args)])
    print('{:35} {}'.format(operation.name + ':', syntax_form))



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utilities for TF-Coder."""

import functools
import itertools
import operator
from typing import List

import numpy as np
import tensorflow as tf
from tf_coder import tensor_limits as limits


# Types of primitive objects.
PRIMITIVE_TYPES = (int, float, bool, str)

# Int tf.DType objects for different sizes.
UINT_DTYPES = (tf.uint32, tf.uint64, tf.uint16, tf.uint8)
INT_DTYPES = (tf.int32, tf.int64, tf.int16, tf.int8) + UINT_DTYPES

# Float tf.DType objects for different sizes.
FLOAT_DTYPES = (tf.float32, tf.float64, tf.float16)

# The prefix for the name of a TensorFlow function.
TF_PREFIX = 'tf.'

# Maps integral dtypes to their min and max values.
INT_DTYPE_MIN_MAX = {
    tf.int8: (-2**7, 2**7 - 1),
    tf.uint8: (0, 2**8 - 1),
    tf.int16: (-2**15, 2**15 - 1),
    tf.uint16: (0, 2**16 - 1),
    tf.int32: (-2**31, 2**31 - 1),
    tf.uint32: (0, 2**32 - 1),
    tf.int64: (-2**63, 2**63 - 1),
    tf.uint64: (0, 2**64 - 1),
}


def get_tf_function(function_name):
  """Returns a TensorFlow function object given its name.

  Args:
    function_name: The string name of the function, e.g., "tf.matmul". Must
      start with "tf.". Nested modules are allowed, e.g., "tf.nn.softmax".

  Returns:
    The function object corresponding to function_name.

  Raises:
    ValueError: If the function name does not start with "tf.", or the function
      could not be found.
  """
  if not function_name.startswith(TF_PREFIX):
    raise ValueError('get_tf_function() called with function {}, which does '
                     'not start with "tf.".'.format(function_name))
  function_name_without_prefix = function_name[len(TF_PREFIX):]
  try:
    tf_function = operator.attrgetter(function_name_without_prefix)(tf)
    if tf_function is None:
      raise ValueError('Could not find TF function {}'.format(function_name))
    return tf_function
  except AttributeError:
    raise ValueError('AttributeError encountered in get_tf_function for name {}'
                     .format(function_name))


def convert_to_tensor(tensor_like):
  """Converts a tensor-like object (e.g., [[1, 2], [3, 4]]) into a tf.Tensor.

  Args:
    tensor_like: A tf.Tensor, tf.SparseTensor, n-dimensional list, or a scalar.

  Returns:
    A tf.Tensor.
  """
  if isinstance(tensor_like, tf.Tensor):
    return tensor_like
  if isinstance(tensor_like, tf.SparseTensor):
    return tf.sparse.reorder(tensor_like)
  return tf.constant(tensor_like)


def num_tensor_elements(tensor):
  """Returns the number of elements in a tensor as an int (primitive)."""
  return int(tf.size(tensor))


def max_tensor_value(tensor):
  """Returns the maximum value in a tensor, as a float (primitive)."""
  return float(tf.reduce_max(tf.cast(tensor, tf.float32)))


def min_tensor_value(tensor):
  """Returns the minimum value in a tensor, as a float (primitive)."""
  return float(tf.reduce_min(tf.cast(tensor, tf.float32)))


def tensor_to_string(tensor, decimals=limits.NUM_DECIMALS):
  """Converts a tensor into a string representation used for equality tests.

  TF-Coder considers two tensors to be equal if and only if their string
  representations (as computed by this function) are equal.

  Args:
    tensor: A Tensor.
    decimals: The number of floating-point decimal places to consider.

  Returns:
    A string representation of the tensor.
  """
  np_array = tensor.numpy()
  if np_array.dtype in [np.float32, np.float64]:
    np_array = np.around(np_array, decimals=decimals)

  # str(np_array.tolist()) is significantly faster than str(np_array).
  return repr(tensor.dtype) + ':' + str(np_array.tolist())


def object_to_string(obj, decimals=limits.NUM_DECIMALS):
  """Converts an object into a string representation used for equality tests.

  TF-Coder considers two objects to be equal if and only if their string
  representations (as computed by this function) are equal.

  Note that two sequences are considered the same if their elements are the
  same, even if one sequence is a list and the other is a tuple.

  Args:
    obj: An object, which could be a Tensor, SparseTensor, TensorFlow dtype,
      primitive (int, float, bool, or string), or sequence (list, tuple, or
      namedtuple) of other such objects.
    decimals: As described in tensor_to_string().

  Returns:
    A string representation of the object.

  Raises:
    ValueError: If `obj` has an unsupported type.
  """
  # Tensors.
  if isinstance(obj, tf.Tensor):
    return tensor_to_string(obj)
  if isinstance(obj, tf.SparseTensor):
    # TODO(kshi): Round float SparseTensors according to `decimals`.
    return str(obj)

  obj_type = type(obj)

  # Primitives and TensorFlow dtypes are handled the same way (with repr()).
  if obj_type in (int, float, bool, str, tf.DType):
    if obj_type == float:
      # Floats must be rounded.
      obj = round(obj, decimals)
    return repr(obj)

  # Sequences (lists, tuples, and namedtuples) of supported objects.
  if obj_type == list or isinstance(obj, tuple):
    return 'seq[' + ', '.join(object_to_string(elem, decimals=decimals)
                              for elem in obj) + ']'

  # All other types are currently unsupported.
  raise ValueError('object_to_string called with unsupported object; type={} '
                   'and str={}.'.format(obj_type, obj))


@functools.lru_cache(maxsize=None)
def generate_partitions(num_elements: int, num_parts: int) -> List[List[int]]:
  """Generates partitions of num_elements into num_parts nonnegative parts.

  Args:
    num_elements: The number of things to permute (a nonnegative integer).
    num_parts: The number of groups to partition into (a positive integer).

  Returns:
    All possible lists of length num_parts, such that the list's elements are
    all nonnegative integers summing to num_elements.

  Raises:
    ValueError: If num_elements is negative, or num_parts is not positive.
  """
  if num_elements < 0:
    raise ValueError('In generate_partitions(), num_elements must be '
                     'nonnegative.')
  if num_parts <= 0:
    raise ValueError('In generate_partitions(), num_parts must be positive.')

  # A list [0, 1, ..., num_elements].
  choices = range(num_elements + 1)

  results = []
  # Choose (num_parts - 1) dividers among the choices, to get num_parts parts.
  for dividers in itertools.combinations_with_replacement(choices,
                                                          num_parts - 1):
    # Add dividers at the first and last choice.
    dividers = [0] + list(dividers) + [num_elements]
    # Pairwise difference between dividers gives the partition.
    results.append([next_divider - divider
                    for divider, next_divider in zip(dividers, dividers[1:])])

  return results



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Bag-of-words description handlers for TF-Coder."""

import pickle
from typing import Dict, List, Optional, Text

from sklearn import ensemble
from sklearn import feature_extraction
from sklearn import naive_bayes
from sklearn import svm
from tf_coder.benchmarks import benchmark as benchmark_module
from tf_coder.datasets.github import data_loader
from tf_coder.datasets.github import tokenizer
from tf_coder.natural_language import description_handler
from tf_coder.value_search import operation_base
from tf_coder.value_search import value_search_settings as settings_module

STOPWORDS = 'english'


def _save_model(filepath, vectorizer, classifiers):
  with open(filepath, 'wb') as f:
    pickle.dump([vectorizer, classifiers], f)


def _load_model(filepath):
  with open(filepath, 'rb') as f:
    vectorizer, classifiers = pickle.load(f)
  return vectorizer, classifiers


class BagOfWordsDescriptionHandler(description_handler.DescriptionHandler):
  """Fit a classifier to estimate the probability of an op's usage."""

  def __init__(self,
               operations: Optional[List[operation_base.Operation]] = None,
               max_features=None,
               make_classifier=None,
               lowercase: bool = True,
               sublinear_tf: bool = True,
               max_num_prioritized: int = 3,
               min_probability: float = 0.5,
               use_docstrings_for_tfidf=False,
               use_docstrings_as_examples=True,
               multiplier: float = 0.75,
               path_prefix: Optional[Text] = None,
               model_filepath=None,
               load: bool = False,
               save: bool = False):
    """Creates a BagOfWordsDescriptionHandler.

    Args:
      operations: A list of operations that the handler should handle.
      max_features: If not None, the max size of the vocabulary to construct in
        the vectorizer.
      make_classifier: A function to produce an untrained classifier, which
        will then be fit to the data.
      lowercase: Whether to lowercase tokens as part of preprocessing.
      sublinear_tf: Whether to use sublinear tfs in the tfidf calculation.
      max_num_prioritized: The maximum number of operations to prioritize.
      min_probability: The minimum score for an operation to be prioritized.
      use_docstrings_for_tfidf: Whether to initialize the tf-idf vectorizer
        using tf operation docstrings.
      use_docstrings_as_examples: Whether to include additional examples, one
        per opperation docstring.
      multiplier: The multiplier applied to an operation's weight if it is
        prioritized.
      path_prefix: The file path prefix from which to load the GitHub data.
      model_filepath: The model filepath to use for saving or loading.
      load: If True, loads a saved model.
      save: If True, pickles the model for later use.
    """
    super(BagOfWordsDescriptionHandler, self).__init__(operations)
    self.path_prefix = path_prefix or data_loader.DEFAULT_DATA_PREFIX
    self.data = data_loader.load_data(self.path_prefix)

    self.max_num_prioritized = max_num_prioritized
    self.min_probability = min_probability
    self.multiplier = multiplier

    self.predict_fns = []
    if load:
      # Load an existing model.
      self.vectorizer, classifiers = _load_model(model_filepath)
      for classifier in classifiers:
        self.predict_fns.append(make_predict_fn(classifier))
      return

    if use_docstrings_as_examples:
      extra_examples = [{  # pylint: disable=g-complex-comprehension
          'docstring': [operation.metadata.docstring],
          'names': [],
          'comments': [],
          'strings': [],
          'tf_functions': [operation.name.split('(')[0]],
      } for operation in self.operations]
      self.data += extra_examples

    # Generate a new model.
    self.vectorizer = feature_extraction.text.TfidfVectorizer(
        stop_words=STOPWORDS,
        max_features=max_features,
        lowercase=lowercase,
        sublinear_tf=sublinear_tf,
    )
    contexts_list = [
        data_loader.get_full_context(example)
        for example in self.data]
    docstrings = [operation.metadata.docstring
                  for operation in self.operations]
    if use_docstrings_for_tfidf:
      self.vectorizer.fit(docstrings)
    else:
      self.vectorizer.fit(contexts_list)
    contexts_features = self.vectorizer.transform(contexts_list)

    self.classifiers = []
    for op_name in self.all_names:
      op = op_name.split('(')[0]
      targets = [data_loader.uses_operation(example, op)
                 for example in self.data]

      classifier = make_classifier()
      try:
        classifier.fit(contexts_features, targets)
      except ValueError:
        # For classes without positive examples, some classifiers cannot be fit.
        classifier = None

      self.classifiers.append(classifier)
      self.predict_fns.append(make_predict_fn(classifier))

    if save:
      _save_model(model_filepath, self.vectorizer, self.classifiers)

  def score_description(self, description: Text) -> Dict[Text, float]:
    description_as_context = ' '.join(tokenizer.tokenize(description))
    vectorized = self.vectorizer.transform([description_as_context])
    probas = {}
    for op_name, predict_fn in zip(self.all_names, self.predict_fns):
      proba = predict_fn(vectorized)
      probas[op_name] = proba
    return probas

  def get_operation_multipliers(
      self, benchmark: benchmark_module.Benchmark,
      settings: settings_module.Settings) -> Dict[Text, float]:
    """See base class."""
    scores = self.score_description(benchmark.description)

    sorted_names = sorted(self.all_names,
                          key=lambda name: scores[name], reverse=True)
    prioritized_set = set(name
                          for name in sorted_names[:self.max_num_prioritized]
                          if scores[name] >= self.min_probability)

    if settings.printing.prioritized_operations:
      for name in prioritized_set:
        print('BOW handler prioritized {}, score={:.3f}'.format(
            name, scores[name]))
    return dict((name, self.multiplier) for name in prioritized_set)

  def __repr__(self) -> Text:
    """See base class."""
    return ('{}(path_prefix={}, min_probability={}, max_num_prioritized={}, '
            'multiplier={})').format(
                self.__class__.__name__,
                self.path_prefix,
                self.min_probability,
                self.max_num_prioritized,
                self.multiplier,
            )


def make_predict_fn(classifier):
  """Produces a single-example predict_proba function from a classifier."""
  if not classifier:
    return lambda v: 0.0

  def predict(vectorized):
    try:
      return classifier.predict_proba(vectorized)[0, 1]
    except:  # pylint: disable=bare-except
      return float(classifier.predict(vectorized)[0])
  return predict


# pylint: disable=invalid-name
def NaiveBayesDescriptionHandler(alpha=0.25, max_features=None,
                                 sublinear_tf=True, max_num_prioritized=3,
                                 min_probability=0.5):
  return BagOfWordsDescriptionHandler(
      max_features=max_features,
      sublinear_tf=sublinear_tf,
      make_classifier=
      lambda: naive_bayes.MultinomialNB(alpha=alpha, class_prior=(.5, .5)),
      max_num_prioritized=max_num_prioritized,
      min_probability=min_probability,
  )


def LinearSVCDescriptionHandler(C=0.0625, max_features=None,
                                max_num_prioritized=3,
                                min_probability=0.5):
  return BagOfWordsDescriptionHandler(
      max_features=max_features,
      make_classifier=
      lambda: svm.LinearSVC(C=C, dual=True, max_iter=500),
      max_num_prioritized=max_num_prioritized,
      min_probability=min_probability,
  )


def _make_ensemble(classifiers, voting='hard'):
  return ensemble.VotingClassifier(
      [('c' + str(i), classifier)
       for i, classifier in enumerate(classifiers)],
      voting=voting
  )


def EnsembleLinearSVCDescriptionHandler(C=0.0625, max_features=None, n=10):
  return BagOfWordsDescriptionHandler(
      max_features=max_features,
      make_classifier=
      # pylint: disable=g-long-lambda
      lambda: _make_ensemble([svm.LinearSVC(C=C, dual=True, max_iter=500)
                              for _ in range(n)]),
      # pylint: enable=g-long-lambda
  )


def EnsembleNaiveBayesDescriptionHandler(alpha=0.25, max_features=None, n=10):
  return BagOfWordsDescriptionHandler(
      max_features=max_features,
      make_classifier=
      # pylint: disable=g-long-lambda
      lambda: _make_ensemble([
          naive_bayes.MultinomialNB(alpha=alpha, fit_prior=True)
          for _ in range(n)]),
      # pylint: enable=g-long-lambda
  )
# pylint: enable=invalid-name



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for value_search.py."""

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
from tf_coder.value_search import value_search_main


class ValueSearchMainTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('all', 'ALL', True),
      ('stackoverflow_01', 'stackoverflow_01', True),
      ('nonexistent_benchmark', 'no_benchmark_has_this_name', False))
  def test_benchmark_name_validator(self, benchmark_name, expected_result):
    self.assertEqual(
        value_search_main.benchmark_name_validator(benchmark_name),
        expected_result)


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Functions and arguments used in the TF-Coder project."""

import ast
import collections

import tensorflow as tf
from tf_coder import filter_group


FilterGroup = filter_group.FilterGroup


FunctionInfo = collections.namedtuple(
    'FunctionInfo',
    ['name', 'filter_group', 'weight'])


# Weights for leaf nodes in the AST.

# Constants given by the user.
PROVIDED_CONSTANT_WEIGHT = 7

# Ubiquitous constants: 0, 1, -1.
COMMON_CONSTANT_WEIGHT = 8

# A tf.constant() wrapper around an input primitive.
PRIMITIVE_INPUT_AS_TENSOR_WEIGHT = 9

# Int constants meant to be axis values, chosen based on input tensor ranks.
AXIS_CONSTANT_WEIGHT = 14

# Int constants obtained from input/output tensor shapes.
SHAPE_CONSTANT_WEIGHT = 24

# Weight of constructing a tuple with the output shape.
OUTPUT_SHAPE_TUPLE_WEIGHT = 32

# Input variable nodes (in1, in2, etc.).
INPUT_VARIABLE_WEIGHT = 8

# DTypes with weights to add to the pool of constants.
CONSTANT_DTYPES_AND_WEIGHTS = collections.OrderedDict([
    (tf.int32, 8),
    (tf.float32, 8),
    (tf.bool, 8),
    (tf.int64, 16),
])

# Used in value search for custom cast logic.
CAST_OPERATION_NAME = 'tf.cast(x, dtype)'

# Used in value search to convert primitive inputs (e.g., 3) into scalar tensors
# (e.g., tf.constant(3)).
CONSTANT_OPERATION_NAME = 'tf.constant(value)'


# A list of FunctionInfo namedtuples, each describing one function usable by a
# program synthesizer. Each FunctionInfo's name contains the function name along
# with the names of the arguments for that function, in the order given in the
# function's signature. A function may appear multiple times with different
# lists of usable arguments. This list is ordered, so value search will try
# earlier functions before later ones.

# FunctionInfo name format: "tf.module.function(arg_1, arg_2, arg_3='value')"
# means call the function `tf.module.function` with varying inputs `arg_1` and
# `arg_2`, where `arg_3` is fixed and set to the literal constant `'value'`.
TF_FUNCTIONS = [
    FunctionInfo(name='tf.abs(x)',
                 filter_group=FilterGroup.PRIMITIVE_OR_TENSOR_1,
                 weight=40),
    FunctionInfo(name='tf.add(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=28),
    FunctionInfo(name='tf.add_n(inputs)',
                 filter_group=FilterGroup.SEQUENCE_1,
                 weight=44),
    FunctionInfo(name='tf.argmax(input, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=32),
    FunctionInfo(name='tf.argmin(input, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=48),
    FunctionInfo(name='tf.argsort(values, axis, stable=True)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=40),
    FunctionInfo(name=("tf.argsort(values, axis, direction='DESCENDING', "
                       "stable=True)"),
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=48),
    FunctionInfo(name='tf.boolean_mask(tensor, mask)',
                 filter_group=FilterGroup.TENSOR_BOOLTENSOR_2,
                 weight=28),
    FunctionInfo(name='tf.broadcast_to(input, shape)',
                 # TODO(kshi): BROADCAST_TO_2 is a workaround for b/124261199.
                 # Once that is fixed, the TENSOR_SHAPE filter_group would be
                 # simpler but might be slower (need to measure).
                 filter_group=FilterGroup.BROADCAST_TO_2,
                 weight=44),
    FunctionInfo(name=CAST_OPERATION_NAME,  # 'tf.cast(x, dtype)'.
                 filter_group=FilterGroup.CASTABLE_DTYPE_2,
                 weight=16),
    FunctionInfo(name='tf.clip_by_value(t, clip_value_min, clip_value_max)',
                 filter_group=FilterGroup.CLIP_BY_VALUE_3,
                 weight=44),
    FunctionInfo(name='tf.concat(values, axis)',
                 filter_group=FilterGroup.TENSORSEQUENCE_AXIS_2,
                 weight=36),
    FunctionInfo(name=CONSTANT_OPERATION_NAME,  # 'tf.constant(value)'.
                 filter_group=FilterGroup.NOT_TENSOR_1,
                 weight=23),  # Less weight than cast, accounting for the dtype.
    FunctionInfo(name='tf.constant(value, dtype)',
                 filter_group=FilterGroup.CASTABLE_DTYPE_2,
                 weight=24),
    FunctionInfo(name='tf.divide(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=28),
    FunctionInfo(name='tf.equal(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=24),
    FunctionInfo(name='tf.exp(x)',
                 filter_group=FilterGroup.FLOATTENSOR_1,
                 weight=52),
    FunctionInfo(name='tf.expand_dims(input, axis)',
                 filter_group=FilterGroup.EXPAND_DIMS_2,
                 weight=18),
    FunctionInfo(name='tf.eye(num_rows)',
                 filter_group=FilterGroup.EYE_1,
                 weight=40),
    FunctionInfo(name='tf.eye(num_rows, num_columns)',
                 filter_group=FilterGroup.EYE_ROWS_COLS_2,
                 weight=60),
    FunctionInfo(name='tf.eye(num_rows, dtype)',
                 filter_group=FilterGroup.EYE_ROWS_DTYPE_2,
                 weight=48),
    FunctionInfo(name='tf.fill(dims, value)',
                 filter_group=FilterGroup.SHAPE_PRIMITIVE_2,
                 weight=40),
    FunctionInfo(name='tf.gather(params, indices)',
                 filter_group=FilterGroup.GATHER_2,
                 weight=24),
    FunctionInfo(name='tf.gather(params, indices, axis, batch_dims)',
                 filter_group=FilterGroup.GATHER_4,
                 weight=40),
    FunctionInfo(name='tf.gather_nd(params, indices)',
                 filter_group=FilterGroup.GATHER_ND_2,
                 weight=28),
    FunctionInfo(name='tf.gather_nd(params, indices, batch_dims)',
                 filter_group=FilterGroup.GATHER_ND_3,
                 weight=36),
    FunctionInfo(name='tf.greater(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=24),
    FunctionInfo(name='tf.greater_equal(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=32),
    FunctionInfo(name='tf.math.bincount(arr)',
                 filter_group=FilterGroup.BINCOUNT_1,
                 weight=40),
    FunctionInfo(name='tf.math.ceil(x)',
                 filter_group=FilterGroup.FLOATTENSOR_1,
                 weight=44),
    FunctionInfo(name='tf.math.count_nonzero(input)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=56),
    FunctionInfo(name='tf.math.count_nonzero(input, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=56),
    FunctionInfo(name='tf.math.cumsum(x, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=44),
    FunctionInfo(name='tf.math.cumsum(x, axis, exclusive=True)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=48),
    FunctionInfo(name='tf.math.divide_no_nan(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_FLOAT_BROADCASTABLE_2,
                 weight=52),
    FunctionInfo(name='tf.math.floor(x)',
                 filter_group=FilterGroup.FLOATTENSOR_1,
                 weight=44),
    FunctionInfo(name='tf.math.log(x)',
                 filter_group=FilterGroup.FLOATTENSOR_1,
                 weight=52),
    FunctionInfo(name='tf.math.logical_and(x, y)',
                 filter_group=FilterGroup.BOOLTENSOR_BROADCASTABLE_2,
                 weight=44),
    FunctionInfo(name='tf.math.logical_not(x)',
                 filter_group=FilterGroup.BOOLTENSOR_1,
                 weight=48),
    FunctionInfo(name='tf.math.logical_or(x, y)',
                 filter_group=FilterGroup.BOOLTENSOR_BROADCASTABLE_2,
                 weight=44),
    FunctionInfo(name='tf.math.logical_xor(x, y)',
                 filter_group=FilterGroup.BOOLTENSOR_BROADCASTABLE_2,
                 weight=60),
    FunctionInfo(name='tf.math.negative(x)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=48),
    FunctionInfo(name='tf.math.reciprocal(x)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=52),
    FunctionInfo(name='tf.math.reciprocal_no_nan(x)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=60),
    FunctionInfo(name='tf.math.segment_max(data, segment_ids)',
                 filter_group=FilterGroup.SEGMENT_OPERATION_2,
                 weight=40),
    FunctionInfo(name='tf.math.segment_mean(data, segment_ids)',
                 filter_group=FilterGroup.SEGMENT_OPERATION_2,
                 weight=56),
    FunctionInfo(name='tf.math.segment_min(data, segment_ids)',
                 filter_group=FilterGroup.SEGMENT_OPERATION_2,
                 weight=48),
    FunctionInfo(name='tf.math.segment_prod(data, segment_ids)',
                 filter_group=FilterGroup.SEGMENT_OPERATION_2,
                 weight=60),
    FunctionInfo(name='tf.math.segment_sum(data, segment_ids)',
                 filter_group=FilterGroup.SEGMENT_OPERATION_2,
                 weight=40),
    FunctionInfo(name='tf.math.squared_difference(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=52),
    FunctionInfo(name='tf.math.top_k(input, k)',
                 filter_group=FilterGroup.TOP_K_2,
                 weight=48),
    FunctionInfo(name=('tf.math.unsorted_segment_max(data, segment_ids, '
                       'num_segments)'),
                 filter_group=FilterGroup.UNSORTED_SEGMENT_OPERATION_3,
                 weight=40),
    FunctionInfo(name=('tf.math.unsorted_segment_mean(data, segment_ids, '
                       'num_segments)'),
                 filter_group=FilterGroup.UNSORTED_SEGMENT_OPERATION_3,
                 weight=56),
    FunctionInfo(name=('tf.math.unsorted_segment_min(data, segment_ids, '
                       'num_segments)'),
                 filter_group=FilterGroup.UNSORTED_SEGMENT_OPERATION_3,
                 weight=48),
    FunctionInfo(name=('tf.math.unsorted_segment_prod(data, segment_ids, '
                       'num_segments)'),
                 filter_group=FilterGroup.UNSORTED_SEGMENT_OPERATION_3,
                 weight=60),
    FunctionInfo(name=('tf.math.unsorted_segment_sum(data, segment_ids, '
                       'num_segments)'),
                 filter_group=FilterGroup.UNSORTED_SEGMENT_OPERATION_3,
                 weight=40),
    FunctionInfo(name='tf.matmul(a, b)',
                 filter_group=FilterGroup.MATMUL_2,
                 weight=24),
    FunctionInfo(name='tf.maximum(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=24),
    FunctionInfo(name='tf.minimum(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=32),
    FunctionInfo(name='tf.multiply(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=24),
    FunctionInfo(name='tf.not_equal(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=44),
    FunctionInfo(name='tf.one_hot(indices, depth)',
                 filter_group=FilterGroup.ONE_HOT_2,
                 weight=28),
    FunctionInfo(name='tf.ones(shape)',
                 filter_group=FilterGroup.SHAPE_1,
                 weight=44),
    FunctionInfo(name='tf.ones_like(input)',
                 filter_group=FilterGroup.TENSOR_1,
                 weight=36),
    FunctionInfo(name="tf.pad(tensor, paddings, mode='CONSTANT')",
                 filter_group=FilterGroup.PAD_2,
                 weight=40),
    FunctionInfo(name=("tf.pad(tensor, paddings, mode='CONSTANT', "
                       "constant_values)"),
                 filter_group=FilterGroup.PAD_3,
                 weight=52),
    FunctionInfo(name="tf.pad(tensor, paddings, mode='REFLECT')",
                 filter_group=FilterGroup.PAD_2,
                 weight=60),
    FunctionInfo(name="tf.pad(tensor, paddings, mode='SYMMETRIC')",
                 filter_group=FilterGroup.PAD_2,
                 weight=60),
    # This behaves like tf.range(limit), but the function signature has `start`
    # as the first argument, and it's not a keyword argument. Fortunately, when
    # we print expressions using this function, it'll be like 'tf.range(1)', and
    # not 'tf.range(start=1)' which would be technically correct but confusing.
    FunctionInfo(name='tf.range(start)',
                 filter_group=FilterGroup.RANGE_1,
                 weight=28),
    FunctionInfo(name='tf.range(start, limit, delta)',
                 filter_group=FilterGroup.RANGE_3,
                 weight=56),
    FunctionInfo(name='tf.reduce_any(input_tensor, axis)',
                 filter_group=FilterGroup.BOOLTENSOR_AXIS_2,
                 weight=40),
    FunctionInfo(name='tf.reduce_all(input_tensor, axis)',
                 filter_group=FilterGroup.BOOLTENSOR_AXIS_2,
                 weight=44),
    FunctionInfo(name='tf.reduce_max(input_tensor)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=24),
    FunctionInfo(name='tf.reduce_max(input_tensor, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=24),
    FunctionInfo(name='tf.reduce_mean(input_tensor)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=40),
    FunctionInfo(name='tf.reduce_mean(input_tensor, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=40),
    FunctionInfo(name='tf.reduce_min(input_tensor)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=40),
    FunctionInfo(name='tf.reduce_min(input_tensor, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=40),
    FunctionInfo(name='tf.reduce_prod(input_tensor, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=52),
    FunctionInfo(name='tf.reduce_sum(input_tensor)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=24),
    FunctionInfo(name='tf.reduce_sum(input_tensor, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=24),
    FunctionInfo(name='tf.repeat(input, repeats)',
                 filter_group=FilterGroup.REPEAT_2,
                 weight=56),
    FunctionInfo(name='tf.repeat(input, repeats, axis)',
                 filter_group=FilterGroup.REPEAT_3,
                 weight=52),
    FunctionInfo(name='tf.reshape(tensor, shape)',
                 filter_group=FilterGroup.TENSOR_SHAPE_2,
                 weight=28),
    FunctionInfo(name='tf.reverse(tensor, axis)',
                 filter_group=FilterGroup.TENSOR_AXISSEQUENCE_2,
                 weight=48),
    FunctionInfo(name='tf.roll(input, shift, axis)',
                 filter_group=FilterGroup.ROLL_3,
                 weight=48),
    FunctionInfo(name='tf.round(x)',
                 filter_group=FilterGroup.FLOATTENSOR_1,
                 weight=52),
    FunctionInfo(name='tf.scatter_nd(indices, updates, shape)',
                 filter_group=FilterGroup.SCATTER_ND_3,
                 weight=52),
    FunctionInfo(name="tf.searchsorted(sorted_sequence, values, side='left')",
                 filter_group=FilterGroup.SEARCHSORTED_2,
                 weight=56),
    FunctionInfo(name="tf.searchsorted(sorted_sequence, values, side='right')",
                 filter_group=FilterGroup.SEARCHSORTED_2,
                 weight=56),
    FunctionInfo(name='tf.sequence_mask(lengths)',
                 filter_group=FilterGroup.SEQUENCE_MASK_1,
                 weight=32),
    FunctionInfo(name='tf.sequence_mask(lengths, maxlen)',
                 filter_group=FilterGroup.SEQUENCE_MASK_2,
                 weight=44),
    FunctionInfo(name='tf.shape(input)',
                 filter_group=FilterGroup.TENSOR_1,
                 weight=36),
    FunctionInfo(name='tf.sign(x)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=48),
    FunctionInfo(name='tf.sort(values, axis)',
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=52),
    FunctionInfo(name="tf.sort(values, axis, direction='DESCENDING')",
                 filter_group=FilterGroup.NUMERICTENSOR_AXIS_2,
                 weight=60),
    FunctionInfo(name='tf.sqrt(x)',
                 filter_group=FilterGroup.FLOATTENSOR_1,
                 weight=56),
    FunctionInfo(name='tf.square(x)',
                 filter_group=FilterGroup.NUMERICTENSOR_1,
                 weight=28),
    FunctionInfo(name='tf.squeeze(input)',
                 filter_group=FilterGroup.SQUEEZE_1,
                 # More weight than tf.squeeze(input, axis), since this version
                 # of squeeze is very prone to false positives (squeezing
                 # unintended axes when those axes have length 1). Realistically
                 # this op only helps when we intend to squeeze multiple axes.
                 weight=24 + AXIS_CONSTANT_WEIGHT),
    FunctionInfo(name='tf.squeeze(input, axis)',
                 filter_group=FilterGroup.SQUEEZE_2,
                 weight=23),  # Less weight than tf.reduce_max(input, axis).
    FunctionInfo(name='tf.stack(values, axis)',
                 filter_group=FilterGroup.TENSORSEQUENCE_AXIS_2,
                 weight=36),
    FunctionInfo(name='tf.subtract(x, y)',
                 filter_group=FilterGroup.SAME_DTYPE_NUMERIC_BROADCASTABLE_2,
                 weight=28),
    FunctionInfo(name='tf.tensor_scatter_nd_update(tensor, indices, updates)',
                 filter_group=FilterGroup.TENSOR_SCATTER_ND_UPDATE_3,
                 weight=44),
    FunctionInfo(name='tf.tensordot(a, b, axes)',
                 filter_group=FilterGroup.TENSORDOT_3,
                 weight=24),
    FunctionInfo(name='tf.tile(input, multiples)',
                 filter_group=FilterGroup.TILE_2,
                 weight=28),
    FunctionInfo(name='tf.transpose(a)',
                 filter_group=FilterGroup.TENSOR_1,
                 weight=24),
    FunctionInfo(name='tf.transpose(a, perm)',
                 filter_group=FilterGroup.TRANSPOSE_2,
                 weight=44),
    FunctionInfo(name='tf.unique_with_counts(x)',
                 filter_group=FilterGroup.TENSOR_1D_1,
                 weight=48),
    FunctionInfo(name='tf.unstack(value, axis)',
                 filter_group=FilterGroup.TENSOR_AXIS_2,
                 weight=40),
    FunctionInfo(name='tf.where(condition)',
                 filter_group=FilterGroup.TENSOR_1,
                 weight=24),
    FunctionInfo(name='tf.where(condition, x, y)',
                 filter_group=FilterGroup.WHERE_3,
                 weight=24),
    FunctionInfo(name='tf.zeros(shape)',
                 filter_group=FilterGroup.SHAPE_1,
                 weight=40),
    FunctionInfo(name='tf.zeros_like(input)',
                 filter_group=FilterGroup.TENSOR_1,
                 weight=32),
]

# Operations for manipulating SparseTensors.
SPARSE_FUNCTIONS = [
    FunctionInfo(name='tf.SparseTensor(indices, values, dense_shape)',
                 filter_group=FilterGroup.SPARSETENSOR_3,
                 weight=20),
    FunctionInfo(name='tf.sparse.add(a, b)',
                 filter_group=FilterGroup.SAME_SHAPE_ONE_SPARSE_2,
                 weight=32),
    FunctionInfo(name='tf.sparse.concat(axis, sp_inputs)',
                 filter_group=FilterGroup.AXIS_SPARSESEQUENCE_2,
                 weight=40),
    FunctionInfo(name='tf.sparse.expand_dims(sp_input, axis)',
                 filter_group=FilterGroup.SPARSE_AXIS_2,
                 weight=24),
    FunctionInfo(name='tf.sparse.from_dense(tensor)',
                 filter_group=FilterGroup.TENSOR_1,
                 weight=20),
    FunctionInfo(name='tf.sparse.maximum(sp_a, sp_b)',
                 filter_group=FilterGroup.SAME_SHAPE_BOTH_SPARSE_2,
                 weight=32),
    FunctionInfo(name='tf.sparse.minimum(sp_a, sp_b)',
                 filter_group=FilterGroup.SAME_SHAPE_BOTH_SPARSE_2,
                 weight=40),
    FunctionInfo(name='tf.sparse.reduce_max(sp_input, axis, output_is_sparse)',
                 filter_group=FilterGroup.SPARSE_AXIS_BOOL_3,
                 weight=28),
    FunctionInfo(name='tf.sparse.reduce_sum(sp_input, axis, output_is_sparse)',
                 filter_group=FilterGroup.SPARSE_AXIS_BOOL_3,
                 weight=28),
    FunctionInfo(name='tf.sparse.reset_shape(sp_input)',
                 filter_group=FilterGroup.SPARSE_1,
                 weight=40),
    FunctionInfo(name='tf.sparse.reshape(sp_input, shape)',
                 filter_group=FilterGroup.SPARSE_SHAPE_2,
                 weight=40),
    FunctionInfo(name='tf.sparse.retain(sp_input, to_retain)',
                 filter_group=FilterGroup.SPARSE_RETAIN_2,
                 weight=36),
    FunctionInfo(name='tf.sparse.slice(sp_input, start, size)',
                 filter_group=FilterGroup.SPARSE_SLICE_3,
                 weight=32),
    FunctionInfo(name='tf.sparse.split(sp_input, num_split, axis)',
                 filter_group=FilterGroup.SPARSE_INT_AXIS_3,
                 weight=32),
    FunctionInfo(name='tf.sparse.to_dense(sp_input)',
                 filter_group=FilterGroup.SPARSE_1,
                 weight=20),
    FunctionInfo(name='tf.sparse.to_dense(sp_input, default_value)',
                 filter_group=FilterGroup.SPARSE_PRIMITIVE_2,
                 weight=36),
    FunctionInfo(name='tf.sparse.to_indicator(sp_input, vocab_size)',
                 filter_group=FilterGroup.SPARSE_TO_INDICATOR_2,
                 weight=44),
    FunctionInfo(name='tf.sparse.transpose(sp_input)',
                 filter_group=FilterGroup.SPARSE_1,
                 weight=36),
    FunctionInfo(name='tf.sparse.transpose(sp_input, perm)',
                 filter_group=FilterGroup.SPARSE_TRANSPOSE_2,
                 weight=56),
]

# A list of operation names that require filtering for value search to work at
# all, i.e., avoid segfaults and huge memory usage. Only relevant for paper
# experiments that turn off filtering for operations not listed here. These
# lists may become stale when we add support for more operations.
REQUIRES_FILTERING = [
    # Potentially huge memory usage.
    'tf.broadcast_to(input, shape)',
    'tf.eye(num_rows)',
    'tf.eye(num_rows, num_columns)',
    'tf.eye(num_rows, dtype)',
    'tf.fill(dims, value)',
    'tf.math.bincount(arr)',
    'tf.math.unsorted_segment_max(data, segment_ids, num_segments)',
    'tf.math.unsorted_segment_mean(data, segment_ids, num_segments)',
    'tf.math.unsorted_segment_min(data, segment_ids, num_segments)',
    'tf.math.unsorted_segment_prod(data, segment_ids, num_segments)',
    'tf.math.unsorted_segment_sum(data, segment_ids, num_segments)',
    'tf.one_hot(indices, depth)',
    'tf.ones(shape)',
    "tf.pad(tensor, paddings, mode='CONSTANT')",
    "tf.pad(tensor, paddings, mode='CONSTANT', constant_values)",
    "tf.pad(tensor, paddings, mode='REFLECT')",
    "tf.pad(tensor, paddings, mode='SYMMETRIC')",
    'tf.range(start)',
    'tf.range(start, limit, delta)',
    'tf.sequence_mask(lengths)',
    'tf.sequence_mask(lengths, maxlen)',
    'tf.tile(input, multiples)',
    'tf.zeros(shape)',

    # Avoid segfaults.
    'tf.cast(x, dtype)',
    'tf.constant(value, dtype)',
    'tf.gather(params, indices, axis, batch_dims)',
    'tf.math.segment_max(data, segment_ids)',
    'tf.math.segment_mean(data, segment_ids)',
    'tf.math.segment_min(data, segment_ids)',
    'tf.math.segment_prod(data, segment_ids)',
    'tf.math.segment_sum(data, segment_ids)',
    'tf.reshape(tensor, shape)',
    'tf.squeeze(input, axis)',
    'tf.sparse.to_indicator(sp_input, vocab_size)',
    'tf.sparse.reshape(sp_input, shape)',
    'tf.sparse.slice(sp_input, start, size)',
    'tf.sparse.split(sp_input, num_split, axis)',
]


def parse_function_info_name(function_info):
  """Takes a FunctionInfo and returns (function_name, list_of_args).

  Args:
    function_info: A FunctionInfo namedtuple.

  Returns:
    A tuple (function_name, list_of_args, constant_kwargs), where function_name
    is a string, list_of_args is a list of strings, and constant_kwargs is a
    dict mapping argument names to their constant literal values. For example,
    if the FunctionInfo's name is 'tf.foo.bar(x, axis, baz=True)', then
    this function would return ('tf.foo.bar', ['x', 'axis'], {'baz': True}).

  Raises:
    ValueError: If the FunctionInfo's name is not properly formatted.
  """
  name = function_info.name

  if name.count('(') != 1:
    raise ValueError("The FunctionInfo's name must have exactly one open "
                     "parenthesis.")
  if name.count(')') != 1 or name[-1] != ')':
    raise ValueError("The FunctionInfo's name must have exactly one close "
                     "parenthesis, at the end of the name.")

  open_paren = name.index('(')
  close_paren = name.index(')')
  function_name = name[ : open_paren]
  arg_list = name[open_paren + 1 : close_paren]
  split_by_comma = [arg.strip() for arg in arg_list.split(',')]
  list_of_args = []
  constant_kwargs = collections.OrderedDict()
  for part in split_by_comma:
    if '=' in part:
      kwarg_name, literal_as_string = [x.strip() for x in part.split('=')]
      constant_kwargs[kwarg_name] = ast.literal_eval(literal_as_string)
    else:
      list_of_args.append(part)
  return function_name, list_of_args, constant_kwargs



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for all_operations.py."""

from absl.testing import absltest
from absl.testing import parameterized
from tf_coder import tf_functions
from tf_coder.value_search import all_operations
from tf_coder.value_search import operation_base


class AllOperationsTest(parameterized.TestCase):

  def test_get_python_operations(self):
    operations = all_operations.get_python_operations()
    self.assertTrue(all(isinstance(operation, operation_base.Operation)
                        for operation in operations))
    self.assertTrue(any(operation.name == 'IndexingOperation'
                        for operation in operations))

  def test_get_tf_operations(self):
    operations = all_operations.get_tf_operations()
    self.assertLen(operations, len(tf_functions.TF_FUNCTIONS))
    self.assertTrue(any(operation.name == 'tf.add(x, y)'
                        for operation in operations))
    self.assertFalse(any(operation.name == 'tf.sparse.add(a, b)'
                         for operation in operations))

  def test_get_sparse_operations(self):
    operations = all_operations.get_sparse_operations()
    self.assertLen(operations, len(tf_functions.SPARSE_FUNCTIONS))
    self.assertFalse(any(operation.name == 'tf.add(x, y)'
                         for operation in operations))
    self.assertTrue(any(operation.name == 'tf.sparse.add(a, b)'
                        for operation in operations))

  def test_get_operations_correct_type(self):
    operations = all_operations.get_operations(include_sparse_operations=True)
    self.assertTrue(all(isinstance(element, operation_base.Operation)
                        for element in operations))

  @parameterized.named_parameters(
      ('with_sparse', True,
       len(tf_functions.TF_FUNCTIONS) + len(tf_functions.SPARSE_FUNCTIONS) +
       len(all_operations.get_python_operations())),
      ('without_sparse', False,
       len(tf_functions.TF_FUNCTIONS) +
       len(all_operations.get_python_operations())))
  def test_get_operations_correct_cardinality(
      self, include_sparse_operations, expected_cardinality):
    operations = all_operations.get_operations(
        include_sparse_operations=include_sparse_operations)
    self.assertLen(operations, expected_cardinality)

  @parameterized.named_parameters(
      ('indexing', 'IndexingOperation', False),
      ('slicing_axis_0_both', 'SlicingAxis0BothOperation', False),
      ('tf_add', 'tf.add(x, y)', False),
      ('tf_cast', 'tf.cast(x, dtype)', False),
      ('tf_sparse_expand_dims', 'tf.sparse.expand_dims(sp_input, axis)', True))
  def test_get_operations_includes_expected(self, name, is_sparse):
    for include_sparse_operations in [True, False]:
      operations = all_operations.get_operations(
          include_sparse_operations=include_sparse_operations)
      should_be_included = include_sparse_operations or not is_sparse
      self.assertEqual(any(operation.name == name for operation in operations),
                       should_be_included)

  def test_get_operations_unique_names(self):
    operations = all_operations.get_operations(
        include_sparse_operations=True)
    names_set = {operation.name for operation in operations}
    self.assertLen(names_set, len(operations))

  def test_get_operations_all_have_docstrings(self):
    operations = all_operations.get_operations(include_sparse_operations=True)
    self.assertTrue(all(operation.metadata.docstring
                        for operation in operations))

  def test_get_operations_consistent_order(self):
    operations_1 = all_operations.get_operations(include_sparse_operations=True)
    operations_2 = all_operations.get_operations(include_sparse_operations=True)
    self.assertEqual([op.name for op in operations_1],
                     [op.name for op in operations_2])

  def test_find_operation_with_name(self):
    operation = all_operations.find_operation_with_name('tf.add(x, y)')
    self.assertEqual(operation.name, 'tf.add(x, y)')

    operation = all_operations.find_operation_with_name(
        'tf.add(x, y)', operation_list=all_operations.get_tf_operations())
    self.assertEqual(operation.name, 'tf.add(x, y)')

  def test_find_operation_with_name_raises_if_not_found(self):
    with self.assertRaises(ValueError):
      all_operations.find_operation_with_name('bad name')

    with self.assertRaises(ValueError):
      all_operations.find_operation_with_name(
          'tf.add(x, y)', operation_list=all_operations.get_python_operations())


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for operation_base.py."""

import timeit

from absl.testing import absltest
from tf_coder.value_search import filtered_values_cache
from tf_coder.value_search import operation_base
from tf_coder.value_search import operation_statistics
from tf_coder.value_search import value
from tf_coder.value_search import value_search_settings as settings_module


class StrangeAdditionOperation(operation_base.Operation):
  """A simple Operation subclass for testing purposes.

  This strange Operation can compute X + Y, but only if X < Y, and one of the
  following holds:

    * X is divisible by 2 and Y is divisible by 3, or
    * X is divisible by 4 and Y is divisible by 5.
  """

  def __init__(self):
    metadata = operation_base.OperationMetadata(docstring='test docstring')
    super(StrangeAdditionOperation, self).__init__(2, weight=5,
                                                   metadata=metadata)
    self.add_value_filters([lambda arg_value: arg_value.value % 2 == 0,
                            lambda arg_value: arg_value.value % 3 == 0])
    self.add_value_filters([lambda arg_value: arg_value.value % 4 == 0,
                            lambda arg_value: arg_value.value % 5 == 0])
    self.set_apply_filter(
        lambda arg_values: arg_values[0].value < arg_values[1].value)

  def _compute_name(self):
    return 'strange_addition'

  def apply(self, arg_values, settings):
    """See base class."""
    return value.OperationValue(arg_values[0].value + arg_values[1].value,
                                self, arg_values)

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return '(' + arg_strings[0] + ' + ' + arg_strings[1] + ')'


def _value(wrapped_value):
  """A simple utility to create Value objects."""
  return value.ConstantValue(wrapped_value)


class OperationBaseTest(absltest.TestCase):

  def setUp(self):
    super(OperationBaseTest, self).setUp()
    self.operation = StrangeAdditionOperation()
    self.settings = settings_module.default_settings()

  def test_name(self):
    self.assertEqual(self.operation.name, 'strange_addition')
    self.assertEqual(self.operation.name, 'strange_addition')  # Cached.

  def test_metadata(self):
    self.assertEqual(self.operation.metadata.docstring, 'test docstring')

  def test_add_value_filters_raises_for_wrong_number_of_filters(self):
    with self.assertRaises(ValueError):
      self.operation.add_value_filters([lambda _: True])

  def test_apply(self):
    result_value = self.operation.apply([_value(2), _value(9)], self.settings)
    self.assertEqual(result_value.value, 11)

  def test_enumerate_values_with_weight(self):
    values_by_weight = [
        [],  # Weight 0.
        [_value(1), _value(4), _value(9), _value(15)],  # Weight 1.
        [_value(2), _value(6), _value(20), _value(60)],  # Weight 2.
        [_value(10), _value(12)],  # Weight 3.
    ]
    statistics = operation_statistics.OperationStatistics()

    filter_cache = filtered_values_cache.FilteredValuesCache()
    actual_results = self.operation.enumerate_values_with_weight(
        9, values_by_weight, filter_cache,
        end_time=float('inf'), settings=self.settings, statistics=statistics)

    # The operation itself has weight 5. We can get a total weight of 9 by
    # adding arguments of weight 1 and 3, or 2 and 2.
    expected_results = [
        # Weight 1 + weight 3.

        # X divisible by 2, Y divisible by 3, and X < Y.
        _value(16),  # 4 + 12.

        # X divisible by 4, Y divisible by 5, and X < Y.
        _value(14),  # 4 + 10.

        ################################################

        # Weight 2 + weight 2.

        # X divisible by 2, Y divisible by 3, and X < Y.
        _value(8),  # 2 + 6.
        _value(62),  # 2 + 60.
        _value(66),  # 6 + 60.
        _value(80),  # 20 + 60.

        # X divisible by 4, Y divisible by 5, and X < Y.
        _value(80),  # 20 + 60.

        ################################################

        # Weight 3 + weight 1.

        # X divisible by 2, Y divisible by 3, and X < Y.
        _value(25),  # 10 + 15.
        _value(27),  # 12 + 15.

        # X divisible by 4, Y divisible by 5, and X < Y.
        _value(27),  # 12 + 15.
    ]
    self.assertCountEqual(actual_results, expected_results)

    self.assertEqual(statistics.total_apply_count, 10)
    self.assertEqual(statistics.operation_apply_successes,
                     {'strange_addition': 10})

  def test_enumerate_values_with_weight_with_immediate_timeout(self):
    values_by_weight = [
        [],  # Weight 0.
        [_value(1), _value(4), _value(9), _value(15)],  # Weight 1.
        [_value(2), _value(6), _value(20), _value(60)],  # Weight 2.
        [_value(10), _value(12)],  # Weight 3.
    ]
    filter_cache = filtered_values_cache.FilteredValuesCache()
    # The time cutoff is before the function starts executing.
    actual_results = self.operation.enumerate_values_with_weight(
        9, values_by_weight, filter_cache,
        end_time=timeit.default_timer() - 0.1, settings=self.settings)
    self.assertEmpty(actual_results)

  def test_reconstruct_expression(self):
    left_term = self.operation.apply([_value(12), _value(34)], self.settings)
    middle_term = self.operation.apply([_value(5), _value(6)], self.settings)
    right_term = self.operation.apply([middle_term, _value(7)], self.settings)
    result_value = self.operation.apply([left_term, right_term], self.settings)
    self.assertEqual(result_value.reconstruct_expression(),
                     '((12 + 34) + ((5 + 6) + 7))')

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tokenization of docstrings, comments, variable names, and string literals."""

import re
import string
from typing import List, Text


def camel_case_split(text: Text) -> List[Text]:
  matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)',
                        text)
  return [m.group(0) for m in matches]


def normalize(text: Text) -> Text:
  """Normalizes text to a canonical form for use in classifiers."""
  # Strip quotes.
  text = text.replace('"', '').replace("'", '')

  # Remove non-alphanumeric characters.
  text = ''.join([
      char if char in string.ascii_letters + string.digits else ' '
      for char in text
  ])
  tokens = text.split()

  # Split camel-case tokens.
  subtokens = []
  for token in tokens:
    subtokens.extend(camel_case_split(token))

  # Lower case.
  return ' '.join(subtokens).lower()


def tokenize(text: Text) -> List[Text]:
  return normalize(text).split()


def tokens_from_text_list(texts: List[Text]) -> List[Text]:
  tokens = []
  for text in texts:
    tokens.extend(normalize(text).split())
  return tokens



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Computes the size of value search's search space.

Found 134 operations.
# operations with arity 0: 0
# operations with arity 1: 37
# operations with arity 2: 76
# operations with arity 3: 20
# operations with arity 4: 1

Num leaf nodes: 13

There are 2.1364033304069425e+18 expressions using exactly 5 ops and 11 nodes.
There are 2.2841608421687201e+18 expressions using at most 5 ops and 11 nodes.
"""

import itertools

from absl import app
from absl import flags
import numpy as np
from tf_coder import tf_coder_utils as utils
from tf_coder.value_search import all_operations


FLAGS = flags.FLAGS

flags.DEFINE_integer('num_ops', 5, 'The target number of operations.')
flags.DEFINE_integer('num_nodes', 11,
                     'The target number of nodes in the expression tree.')
flags.DEFINE_integer('num_leaf_choices', 13,
                     'The number of inputs and constants available.')


def compute_search_space_size(num_ops, num_nodes, num_leaf_choices):
  """Computes and prints the size of the search space.

  This counts the total number of expressions with exactly the given number of
  operations and nodes in the expression tree. Distinct expressions will be
  counted separately even if they evaluate to the same value, unlike in
  TF-Coder's value_search algorithm which does value-based pruning.

  Args:
    num_ops: The target number of operations.
    num_nodes: A target number of nodes in the expression tree.
    num_leaf_choices: The number of distinct inputs and constants available to
      form the leaves of the expression tree.

  Returns:
    The DP table, where dp[i][j] is the answer for i ops and j nodes.
  """
  operations = all_operations.get_operations(include_sparse_operations=True)
  max_arity = max(op.num_args for op in operations)
  arity_counts = [0] * (max_arity + 1)

  print('Found {} operations.'.format(len(operations)))
  for arity in range(max_arity + 1):
    arity_counts[arity] = sum(1 for op in operations if op.num_args == arity)
    print('# operations with arity {}: {}'.format(arity, arity_counts[arity]))
  print('\nNum leaf nodes: {}'.format(num_leaf_choices))

  # dp[i][j] = the number of expressions using exactly i ops and j nodes.
  dp = np.zeros((num_ops + 1, num_nodes + 1))

  # The only expressions using 0 ops are single-node leaves.
  dp[0][1] = num_leaf_choices

  for ops in range(1, num_ops + 1):
    for nodes in range(1, num_nodes + 1):
      # The running total number of ways to satisfy # ops and # nodes.
      total = 0

      for arity in range(1, max_arity + 1):
        # The running total number of ways to fill the arguments.
        args_total = 0

        # The ways to allocate remaining ops and nodes to the arguments.
        ops_partitions = utils.generate_partitions(num_elements=ops - 1,
                                                   num_parts=arity)
        nodes_partitions = utils.generate_partitions(num_elements=nodes - 1,
                                                     num_parts=arity)

        for ops_partition, nodes_partition in itertools.product(
            ops_partitions, nodes_partitions):
          # The i-th argument must have ops_partition[i] ops and
          # nodes_partition[i] nodes. Look up the number of ways in the DP
          # table.
          args_total += np.prod([dp[ops_partition[i]][nodes_partition[i]]
                                 for i in range(arity)])

        # There are arity_counts[arity] choices for the outermost operation.
        total += args_total * arity_counts[arity]

      dp[ops][nodes] = total

  print('\nThere are {} expressions using exactly {} ops and {} nodes.'.format(
      dp[num_ops][num_nodes], num_ops, num_nodes))
  print('There are {} expressions using at most {} ops and {} nodes.'.format(
      np.sum(dp), num_ops, num_nodes))
  return dp


def main(unused_argv):
  compute_search_space_size(num_ops=FLAGS.num_ops,
                            num_nodes=FLAGS.num_nodes,
                            num_leaf_choices=FLAGS.num_leaf_choices)


if __name__ == '__main__':
  app.run(main)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Simple benchmarks that TF-Coder should always solve quickly."""

# Avoid wrapping URLs and target programs to ease clicking and copying.
# pylint: disable=line-too-long

import tensorflow as tf
from tf_coder.benchmarks import benchmark


def simple_broadcasted_add():
  """Adding two tensors with broadcasting."""
  examples = [
      benchmark.Example(
          inputs=[
              [3, 4, 5],
              [10, 20, 30],
          ],
          output=[[13, 14, 15], [23, 24, 25], [33, 34, 35]],
      ),
  ]
  constants = []
  description = 'Add two tensors with broadcasting'
  target_program = 'tf.add(in1, tf.expand_dims(in2, 1))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_broadcasted_add')


def simple_with_input_names():
  """A benchmark using names for input tensors."""
  examples = [
      benchmark.Example(
          inputs={
              'tensor_x': [3, 4, 5],
              'tensor_y': [10, 20, 30],
          },
          output=[[13, 14, 15], [23, 24, 25], [33, 34, 35]],
      ),
  ]
  constants = []
  description = 'Add two tensors with broadcasting'
  target_program = 'tf.add(tensor_x, tf.expand_dims(tensor_y, 1))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_with_input_names')


def simple_cast():
  """Casting an int tensor into a float tensor."""
  examples = [
      benchmark.Example(
          inputs=[
              [12, 34, 56]
          ],
          output=[12.0, 34.0, 56.0],
      ),
  ]
  constants = []
  description = 'Cast an int tensor into a float tensor'
  target_program = 'tf.cast(in1, tf.float32)'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_cast')


def simple_index():
  """Indexing within a tensor."""
  examples = [
      benchmark.Example(
          inputs=[
              [12, 34, 56, 78],
              -2,
          ],
          output=56,
      ),
  ]
  constants = []
  description = 'Index into a tensor'
  target_program = 'in1[in2]'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_index')


def simple_slice():
  """Slicing within a tensor."""
  examples = [
      benchmark.Example(
          inputs=[
              [[12, 34, 56, 78], [-1, -2, -3, -4]],
              -1,
          ],
          output=[[34, 56], [-2, -3]],
      ),
  ]
  constants = []
  description = 'Slice a tensor'
  target_program = 'in1[:, 1:in2]'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_slice')


def simple_sparse_add():
  """Adding a sparse and a dense tensor to produce a sparse tensor."""
  examples = [
      benchmark.Example(
          inputs=[
              tf.SparseTensor(indices=[[0, 0], [0, 1]],
                              values=[12, 34],
                              dense_shape=[2, 2]),
              [[-3, 0], [-5, 0]],
          ],
          output=tf.SparseTensor(indices=[[0, 0], [0, 1], [1, 0]],
                                 values=[9, 34, -5],
                                 dense_shape=[2, 2]),
      ),
  ]
  constants = []
  description = 'Add sparse tensor with dense tensor'
  target_program = 'tf.sparse.add(in1, tf.sparse.from_dense(in2))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_sparse_add')


def simple_add_big_tensors():
  """Broadcasted addition of "big" tensors."""
  examples = [
      benchmark.Example(
          inputs=[
              list(range(100)),
              [1000, 2000, 3000, 4000, 5000],
          ],
          output=[list(range(start, start + 100))
                  for start in range(1000, 5001, 1000)]
      ),
  ]
  constants = []
  description = 'Add two tensors'
  target_program = 'tf.add(in1, tf.expand_dims(in2, 1))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_add_big_tensors')


def simple_using_constant():
  """A benchmark that requires a user-provided constant."""
  examples = [
      benchmark.Example(
          inputs=[
              [1, 2, 3],
          ],
          output=[101, 102, 103]
      ),
  ]
  constants = [100]
  description = 'Add 100 to every element'
  target_program = 'tf.add(in1, tf.constant(100))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_using_constant')


def simple_using_output_shape():
  """A benchmark that requires a constant obtained from the output shape."""
  examples = [
      benchmark.Example(
          inputs=[
              tf.constant(7.0),
          ],
          output=[[7.0, 0.0, 0.0, 0.0, 0.0],
                  [0.0, 7.0, 0.0, 0.0, 0.0],
                  [0.0, 0.0, 7.0, 0.0, 0.0],
                  [0.0, 0.0, 0.0, 7.0, 0.0],
                  [0.0, 0.0, 0.0, 0.0, 7.0]]
      ),
  ]
  constants = []
  description = 'Multiply with the identity matrix'
  target_program = 'tf.multiply(in1, tf.eye(5))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_using_output_shape')


def simple_using_output_shape_tuple():
  """A benchmark that requires the output shape as a constant."""
  examples = [
      benchmark.Example(
          inputs=[],
          output=tf.zeros((2, 3, 4, 5)),
      ),
  ]
  constants = []
  description = 'Construct a 4D zeros tensor'
  target_program = 'tf.zeros((2, 3, 4, 5))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_using_output_shape_tuple')


def simple_using_boolean_constant():
  """A benchmark that requires using a boolean constant."""
  examples = [
      benchmark.Example(
          inputs=[
              tf.SparseTensor(indices=[[0, 0], [0, 1], [1, 1]],
                              values=[12, 34, 56],
                              dense_shape=[2, 2]),
          ],
          output=tf.SparseTensor(indices=[[0], [1]],
                                 values=[46, 56],
                                 dense_shape=[2]),
      ),
  ]
  constants = []
  description = 'Reduce sum on a sparse tensor'
  target_program = 'tf.sparse.reduce_sum(in1, axis=1, output_is_sparse=True)'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_using_boolean_constant')


def simple_using_constant_kwarg():
  """A benchmark that requires using a constant kwarg."""
  examples = [
      benchmark.Example(
          inputs=[
              [40, 20, 60, 50, 10, 20, 50, 50],
          ],
          output=[2, 3, 6, 7, 0, 1, 5, 4],
      ),
  ]
  constants = []
  description = 'Indices sorted in reverse order'
  target_program = "tf.argsort(in1, axis=0, direction='DESCENDING', stable=True)"
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_using_constant_kwarg')


def simple_using_primitive_input():
  """A benchmark with a Python primitive as an input."""
  examples = [
      benchmark.Example(
          inputs=[
              123,
              tf.constant(45),
          ],
          output=tf.constant(168),
      ),
  ]
  constants = []
  description = 'Add primitive int and scalar int tensor'
  target_program = 'tf.add(in2, tf.constant(in1))'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_using_primitive_input')


def simple_with_many_inputs():
  """A benchmark with many input tensors."""
  examples = [
      benchmark.Example(
          inputs={
              'params': [[[11, 22, 33], [44, 55, 66]]],
              'indices': [[2, 1]],
              'axis': 2,
              'batch_dims': 1,
          },
          output=[[[33, 22], [66, 55]]],
      ),
  ]
  constants = []
  description = 'Use tf.gather'
  target_program = 'tf.gather(params, indices, axis=axis, batch_dims=batch_dims)'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_with_many_inputs')


def simple_output_equals_input_single():
  """A benchmark with the output being equal to the input tensor."""
  examples = [
      benchmark.Example(
          inputs={
              'tens': [10, 20, 30, 40, 50],
          },
          output=[10, 20, 30, 40, 50],
      ),
  ]
  constants = [999]
  description = 'find the right input variable'
  target_program = 'tens'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_output_equals_input_single')


def simple_output_equals_input_multiple():
  """A benchmark with the output being equal to one of several input tensors."""
  examples = [
      benchmark.Example(
          inputs={
              'ones': [1, 2, 3, 4, 5],
              'tens': [10, 20, 30, 40, 50],
              'hundreds': [100, 200, 300],
          },
          output=[10, 20, 30, 40, 50],
      ),
  ]
  constants = [999]
  description = 'find the right input variable'
  target_program = 'tens'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_output_equals_input_multiple')


def simple_output_equals_constant():
  """A benchmark with the output being equal to a constant."""
  examples = [
      benchmark.Example(
          inputs={
              'ones': [1, 2, 3, 4, 5],
              'tens': [10, 20, 30, 40, 50],
              'hundreds': [100, 200, 300],
          },
          output=10,
      ),
  ]
  constants = [1, 10, 100]
  description = 'find the right constant'
  target_program = 'tf.constant(10)'
  source = 'handwritten task'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='simple_output_equals_constant')




# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""A cache for operation filtering."""

import collections
from typing import Callable, Dict, Iterable, List, Optional

from tf_coder.value_search import value


class FilteredValuesCache(object):
  """A cache for operation filtering."""

  def __init__(self) -> None:
    # A map from filter functions and weights to associated values.
    self._cache = collections.defaultdict(dict)  # type: Dict[Callable[[value.Value], bool], Dict[int, List[value.Value]]]

  def filter_values(self,
                    filter_function: Optional[Callable[[value.Value], bool]],
                    weight: int,
                    values: Iterable[value.Value]) -> List[value.Value]:
    """Filters values using a filter function, caching the results.

    Args:
      filter_function: A filter function function (Value -> bool). If None, this
        means skip filtering, e.g., equivalent to filter_function being
        `lambda _: True`, but faster.
      weight: The weight of the candidate values. This will be used as part of
        the cache key.
      values: A list of values with the given weight, to be filtered.

    Returns:
      A list of values that pass the filter. The returned list should not be
      modified.
    """
    cached_weights = self._cache[filter_function]
    if weight in cached_weights:
      return cached_weights[weight]
    else:
      filtered_values = list(filter(filter_function, values)
                             if filter_function is not None
                             else values)
      cached_weights[weight] = filtered_values
      return filtered_values



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Benchmarks used only for testing purposes."""

from tf_coder.benchmarks import benchmark


def test_add():
  """Returns a simple benchmark.Benchmark for an addition task."""
  examples = [
      benchmark.Example(
          inputs=[
              [10],
              [20],
          ],
          output=[30],
      ),
  ]
  constants = [0]
  description = 'Add elementwise'
  target_program = 'tf.add(in1, in2)'
  source = 'test'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='test_add')


def test_cast():
  """Returns a simple benchmark.Benchmark for a casting task."""
  examples = [
      benchmark.Example(
          inputs=[
              [1, 0, 1, 1, 0],
          ],
          output=[True, False, True, True, False],
      ),
  ]
  constants = []
  description = 'cast a tensor'
  target_program = 'tf.cast(in1, tf.bool)'
  source = 'test'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='test_cast')


@benchmark.ignore('Target program is inconsistent with the example.')
def inconsistent_target_program():
  """Returns a benchmark.Benchmark with an inconsistent target program."""
  examples = [
      benchmark.Example(
          inputs=[
              [10],
              [20],
          ],
          output=[40],  # Should be 30.
      ),
  ]
  constants = [0]
  description = 'add elementwise'
  target_program = 'tf.add(in1, in2)'
  source = 'test'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='inconsistent_target_program')


@benchmark.ignore('Intentionally duplicate benchmark.')
def duplicate_test_add():
  """Returns a duplicate benchmark.Benchmark."""
  examples = [
      benchmark.Example(
          inputs=[
              [10],
              [20],
          ],
          output=[30],
      ),
  ]
  constants = [0]
  description = 'Add elementwise'
  target_program = 'tf.add(in1, in2)'
  source = 'test'
  return benchmark.Benchmark(examples=examples,
                             constants=constants,
                             description=description,
                             target_program=target_program,
                             source=source,
                             name='test_add')



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines the Operation objects for functions."""

import re

import funcsigs
import tensorflow as tf
from tf_coder import tensor_limits as limits
from tf_coder import tf_coder_utils
from tf_coder import tf_functions
from tf_coder.value_search import operation_base
from tf_coder.value_search import operation_filtering
from tf_coder.value_search import value


class FunctionOperation(operation_base.Operation):
  """An operation that applies a function to some arguments.

  The arguments must be given in the same order as they appear in the function's
  signature.

  Arguments with default values in the function signature are optional at the
  time of FunctionOperation creation. However, once created, a FunctionOperation
  must be used with all of the arguments provided to its constructor.
  """

  def __init__(self, function_info):
    """Creates a FunctionOperation.

    Args:
      function_info: A tf_functions.FunctionInfo.
    """
    function_name, arg_names, constant_kwargs = (
        tf_functions.parse_function_info_name(function_info))
    self._function_obj = tf_coder_utils.get_tf_function(function_name)
    docstring = self._function_obj.__doc__
    if not docstring:
      print('Warning: could not get docstring for function {}'.format(
          function_name))
      docstring = ''

    # Make sure the function and argument names appear in the docstring. (Args
    # should already appear in the docstring "Args" section though.)
    docstring += '\n' + function_info.name
    # If 'reduce_max' is the function name, make sure 'reduce' and 'max' also
    # appear as separate words. Ditto for argument names as well.
    docstring += '\n' + function_info.name.replace('_', ' ')
    # Upweight the function name (moreso than the argument names).
    function_name_without_tf = re.sub(r'^tf\.', '', function_name)
    docstring += ('\n' + function_name_without_tf) * 4
    if '_' in function_name_without_tf:
      docstring += ('\n' + function_name_without_tf.replace('_', ' ')) * 2

    metadata = operation_base.OperationMetadata(docstring=docstring)
    super(FunctionOperation, self).__init__(
        num_args=len(arg_names), weight=function_info.weight, metadata=metadata)

    self.function_info = function_info
    self.function_name = function_name
    self.arg_names = arg_names
    self.constant_kwargs = constant_kwargs

    self._has_default = {}

    parameters = funcsigs.signature(self._function_obj).parameters
    for arg_name in arg_names:
      param = parameters[arg_name]
      has_default = param.default is not param.empty
      self._has_default[arg_name] = has_default

    operation_filtering.add_filters_to_function_operation(self)

  def _compute_name(self):
    return self.function_info.name

  def _print_warnings(self, arg_values, result_value):
    if isinstance(result_value, tf.Tensor):
      num_elements = tf_coder_utils.num_tensor_elements(result_value)
    elif isinstance(result_value, tf.SparseTensor):
      num_elements = tf_coder_utils.num_tensor_elements(result_value.values)
    else:
      return
    if num_elements > 10 * limits.MAX_TENSOR_ELEMENTS:
      print('Warning: {} produced much-too-large tensor of shape {} and {} '
            'elements.'.format(
                self.name,
                result_value.shape.as_list(),
                num_elements))
      for i, arg_value in enumerate(arg_values):
        if isinstance(arg_value.value, tf.Tensor):
          print('  argument {} has shape {} and {} elements'.format(
              i, arg_value.shape, arg_value.num_elements()))
          if arg_value.num_elements() <= 20:
            print('  argument {} is: {}'.format(i, arg_value.value))
        elif arg_value.is_primitive:
          print('  argument {} is: {}'.format(i, arg_value.value))
        else:
          print('  argument {} has type {}'.format(
              i, type(arg_value.value)))
        print('  argument {} has reconstruction: {}'.format(
            i, arg_value.reconstruct_expression()))

  def apply(self, arg_values, settings):
    """See base class."""
    value_objects = [arg_value.value for arg_value in arg_values]
    arg_dict = dict(zip(self.arg_names, value_objects))
    arg_dict.update(self.constant_kwargs)
    try:
      result_value = self._function_obj(**arg_dict)
    except Exception:  # pylint: disable=broad-except
      return None
    try:
      return value.OperationValue(result_value, self, arg_values)
    except ValueError:
      if settings.printing.tensor_size_warnings:
        self._print_warnings(arg_values, result_value)
      return None

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    arg_strings = list(arg_strings)
    for i, arg_name in enumerate(self.arg_names):
      if self._has_default[arg_name]:
        arg_strings[i] = arg_name + '=' + arg_strings[i]
    for kwarg_name, kwarg_value in self.constant_kwargs.items():
      arg_strings.append('{}={!r}'.format(kwarg_name, kwarg_value))
    return self.function_name + '(' + ', '.join(arg_strings) + ')'



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Utilities for value search."""

from tf_coder import tensor_limits as limits


def check_tensor_size(tensor_value):
  """Returns whether the tensor is an acceptable size for value search."""
  tensor = tensor_value.value
  return (
      # Check the number of elements.
      tensor_value.num_elements() <= limits.MAX_TENSOR_ELEMENTS and

      # Check the maximum length of a dimension.

      # Note that a TensorShape object for zero dimensions does NOT implicitly
      # convert to False! We really want to check `len(tensor.shape) == 0`,
      # which is NOT equivalent to `not tensor.shape` as suggested by the
      # linter and style guide.
      (not len(tensor.shape) or  # pylint: disable=g-explicit-length-test
       max(tensor.shape) <= limits.MAX_DIMENSION_LENGTH) and

      # Check the number of dimensions.
      len(tensor.shape) <= limits.MAX_NUM_DIMENSIONS)


def check_sparse_tensor_size(sparse_tensor_value):
  """Returns whether the tensor is an acceptable size for value search."""
  sparse_tensor = sparse_tensor_value.value
  return (
      # Check the number of non-default elements.
      sparse_tensor_value.num_elements() <= limits.MAX_TENSOR_ELEMENTS and

      # Check the number of dimensions.
      len(sparse_tensor.dense_shape) <= limits.MAX_NUM_DIMENSIONS)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Creates description handlers from strings."""

import collections
import functools
from typing import Callable, Dict, List, Text

from tf_coder.natural_language import bag_of_words_handlers
from tf_coder.natural_language import description_handler
from tf_coder.natural_language import tfidf_handler


# Use lambdas to avoid instantiating handlers until they're used.
# pylint: disable=g-long-lambda
DESCRIPTION_HANDLER_FNS = collections.OrderedDict([
    ('no_change',
     description_handler.NoChangeDescriptionHandler),
    ('function_name',
     description_handler.FunctionNameDescriptionHandler),
    ('tfidf',
     tfidf_handler.TfidfDescriptionHandler),  # Use its defaults.
    ('naive_bayes',
     bag_of_words_handlers.NaiveBayesDescriptionHandler),  # Use its defaults.
    ('svm',
     bag_of_words_handlers.LinearSVCDescriptionHandler),  # Use its defaults.

    ('tfidf_3_0.1',
     functools.partial(tfidf_handler.TfidfDescriptionHandler,
                       max_num_prioritized=3,
                       min_tfidf_score=0.1)),
    ('tfidf_3_0.15',
     functools.partial(tfidf_handler.TfidfDescriptionHandler,
                       max_num_prioritized=3,
                       min_tfidf_score=0.15)),
    ('naive_bayes_3_0.5',
     functools.partial(bag_of_words_handlers.NaiveBayesDescriptionHandler,
                       max_num_prioritized=3,
                       min_probability=0.5)),
    ('naive_bayes_3_0.75',
     functools.partial(bag_of_words_handlers.NaiveBayesDescriptionHandler,
                       max_num_prioritized=3,
                       min_probability=0.75)),

    ('tfidf_5_0.1',
     functools.partial(tfidf_handler.TfidfDescriptionHandler,
                       max_num_prioritized=5,
                       min_tfidf_score=0.1)),
    ('tfidf_5_0.15',
     functools.partial(tfidf_handler.TfidfDescriptionHandler,
                       max_num_prioritized=5,
                       min_tfidf_score=0.15)),
    ('naive_bayes_5_0.5',
     functools.partial(bag_of_words_handlers.NaiveBayesDescriptionHandler,
                       max_num_prioritized=5,
                       min_probability=0.5)),
    ('naive_bayes_5_0.75',
     functools.partial(bag_of_words_handlers.NaiveBayesDescriptionHandler,
                       max_num_prioritized=5,
                       min_probability=0.75)),

    ('tfidf_10_0.1',
     functools.partial(tfidf_handler.TfidfDescriptionHandler,
                       max_num_prioritized=10,
                       min_tfidf_score=0.1)),
    ('tfidf_10_0.15',
     functools.partial(tfidf_handler.TfidfDescriptionHandler,
                       max_num_prioritized=10,
                       min_tfidf_score=0.15)),
    ('naive_bayes_10_0.5',
     functools.partial(bag_of_words_handlers.NaiveBayesDescriptionHandler,
                       max_num_prioritized=10,
                       min_probability=0.5)),
    ('naive_bayes_10_0.75',
     functools.partial(bag_of_words_handlers.NaiveBayesDescriptionHandler,
                       max_num_prioritized=10,
                       min_probability=0.75)),
])  # type: Dict[Text, Callable[[], description_handler.DescriptionHandler]]
# pylint: enable=g-long-lambda


def handler_string_list() -> List[Text]:
  """Returns a list of available handler strings."""
  return list(DESCRIPTION_HANDLER_FNS.keys())


def create_handler(
    handler_string: Text) -> description_handler.DescriptionHandler:
  """Returns a DescriptionHandler corresponding to the given handler string."""
  if handler_string not in DESCRIPTION_HANDLER_FNS:
    raise ValueError('Unknown description handler: {}'
                     .format(handler_string))
  # Evaluate the lambda to get the handler.
  return DESCRIPTION_HANDLER_FNS[handler_string]()



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines Operation objects for Python operators."""

import abc

import six
from tf_coder import tensor_limits as limits
from tf_coder import tf_coder_utils
from tf_coder.value_search import operation_base
from tf_coder.value_search import operation_filtering as filtering
from tf_coder.value_search import value

# Weights for Python operations.
INDEXING_WEIGHT = 32
SLICING_WEIGHT = 36
SINGLETON_TUPLE_CREATION_WEIGHT = 20
PAIR_CREATION_WEIGHT = 20
TRIPLE_CREATION_WEIGHT = 36

# "Docstrings" for Python operations, so they can used for ranking in the same
# way as for TensorFlow operations.

INDEXING_DOCSTRING = """
Indexes axis 0 of a tensor or list using standard Python bracket notation.

Selects an element from the tensor or list according to the index. The index
applies to the first dimension (axis 0). Negative indices are supported,
counting backwards from the end.
"""

INDEXING_AXIS_1_DOCSTRING = """
Indexes axis 1 of a tensor using standard Python bracket notation.

Selects a slice of the tensor according to the index. The index applies to the
second dimension (axis 1). Negative indices are supported, counting backwards
from the end.
"""

SINGLETON_TUPLE_CREATION_DOCSTRING = """
Creates a single-element tuple.

This may be used to create a shape for tensors.
"""

PAIR_CREATION_DOCSTRING = """
Creates a pair (2-tuple).

This may be used to pack multiple tensors into a sequence, or create a shape for
tensors.
"""

TRIPLE_CREATION_DOCSTRING = """
Creates a triple (3-tuple).

This may be used to pack multiple tensors into a sequence, or create a shape for
tensors.
"""

# This is used as a format string for all of the slicing variations.
# `slice_kind` should be 'prefix', 'suffix', or 'range'.
# `axis` should be an integer axis number.
SLICING_DOCSTRING_FORMAT = """
Slices a tensor along one dimension using standard Python notation.

Selects a {slice_kind} of indices along axis {axis} of the tensor.
"""


class IndexingOperation(operation_base.Operation):
  """An indexing operation of the form "arg0[arg1]"."""

  def __init__(self):
    metadata = operation_base.OperationMetadata(docstring=INDEXING_DOCSTRING)
    super(IndexingOperation, self).__init__(
        num_args=2, weight=INDEXING_WEIGHT, metadata=metadata)

    def _sequence_or_tensor(arg_value):
      return (arg_value.is_sequence or
              (arg_value.is_tensor and len(arg_value.shape)))
    self.add_value_filters([_sequence_or_tensor,
                            filtering.INT_OR_INT_TENSOR_FILTER])
    def _check_index(arg_values):
      """Checks that the index is in range for the sequence or tensor."""
      indexable, index = arg_values
      if indexable.is_sequence:
        length = len(indexable.value)
      else:  # indexable is a tensor with at least 1 dimension.
        length = indexable.shape[0]
      return -length <= int(index.value) < length
    self.set_apply_filter(_check_index)

  def apply(self, arg_values, settings):
    """See base class."""
    try:
      return value.OperationValue(arg_values[0].value[int(arg_values[1].value)],
                                  self, arg_values)
    except Exception:  # pylint: disable=broad-except
      return None

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[' + arg_strings[1] + ']'


class IndexingAxis1Operation(operation_base.Operation):
  """An indexing operation of the form "arg0[:, arg1]"."""

  def __init__(self):
    metadata = operation_base.OperationMetadata(
        docstring=INDEXING_AXIS_1_DOCSTRING)
    super(IndexingAxis1Operation, self).__init__(
        num_args=2, weight=INDEXING_WEIGHT, metadata=metadata)

    self.add_value_filters([filtering.get_tensor_min_rank_filter(2),
                            filtering.INT_OR_INT_TENSOR_FILTER])
    def _check_index(arg_values):
      """Checks that the index is in range for the tensor."""
      tensor, index = arg_values
      length = tensor.shape[1]
      return -length <= int(index.value) < length
    self.set_apply_filter(_check_index)

  def apply(self, arg_values, settings):
    """See base class."""
    try:
      return value.OperationValue(
          arg_values[0].value[:, int(arg_values[1].value)], self, arg_values)
    except Exception:  # pylint: disable=broad-except
      return None

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[:, ' + arg_strings[1] + ']'


class SingletonTupleCreationOperation(operation_base.Operation):
  """An operation of the form "(arg0,)".

  The argument must be a primitive or sequence. This should not be used to
  create a single-element sequence of tensors, since ops requiring a list of
  tensors are only meaningful if there are multiple tensors in the list.
  """

  def __init__(self):
    metadata = operation_base.OperationMetadata(
        docstring=SINGLETON_TUPLE_CREATION_DOCSTRING)
    super(SingletonTupleCreationOperation, self).__init__(
        num_args=1, weight=SINGLETON_TUPLE_CREATION_WEIGHT, metadata=metadata)

    def _primitives_or_sequences_filter(arg_value):
      """The result must be a possibly-nested list of primitives."""
      return (arg_value.is_primitive or
              filtering.TENSOR_LIKE_SEQUENCE_FILTER(arg_value))
    self.add_value_filters([_primitives_or_sequences_filter])

  def apply(self, arg_values, settings):
    """See base class."""
    try:
      return value.OperationValue((arg_values[0].value,), self, arg_values)
    except ValueError:
      return None

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return '({},)'.format(arg_strings[0])


class PairCreationOperation(operation_base.Operation):
  """An operation of the form "(arg0, arg1)".

  The arguments must be primitives of the same type, or tensors of the same
  dtype and similar shape, or sequences with the same sequence_dtype and
  sequence_shape.
  """

  def __init__(self):
    metadata = operation_base.OperationMetadata(
        docstring=PAIR_CREATION_DOCSTRING)
    super(PairCreationOperation, self).__init__(
        num_args=2, weight=PAIR_CREATION_WEIGHT, metadata=metadata)

    for primitive_type in tf_coder_utils.PRIMITIVE_TYPES:
      self.add_value_filters(
          [filtering.get_type_filter(primitive_type)] * 2)

    def _tensor_value_filter(arg_value):
      """Only keeps values that are "small" tensors."""
      return (arg_value.is_tensor and
              arg_value.num_elements() * 2 <= limits.MAX_TENSOR_ELEMENTS)
    self.add_value_filters([_tensor_value_filter] * 2)

    self.add_value_filters([filtering.TENSOR_LIKE_SEQUENCE_FILTER] * 2)

    def _apply_filter(arg_values):
      """Ensures dtype and shape compatibility."""
      first, second = arg_values
      if first.is_tensor:  # Implies second is also a tensor.
        if first.dtype != second.dtype:
          return False
        shape_1 = first.shape
        shape_2 = second.shape
        if shape_1 == shape_2:
          return True
        if len(shape_1) != len(shape_2):
          return False
        num_different = sum(len_1 != len_2
                            for len_1, len_2 in zip(shape_1, shape_2))
        return num_different <= 1
      elif first.is_sequence:  # Implies second is also a sequence.
        return (first.sequence_dtype == second.sequence_dtype and
                first.sequence_shape == second.sequence_shape)
      else:  # Implies first and second are same-type primitives.
        return True
    self.set_apply_filter(_apply_filter)

  def apply(self, arg_values, settings):
    """See base class."""
    try:
      return value.OperationValue((arg_values[0].value, arg_values[1].value),
                                  self, arg_values)
    except ValueError:
      return None

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return '(' + ', '.join(arg_strings) + ')'


class TripleCreationOperation(operation_base.Operation):
  """An operation of the form "(arg0, arg1, arg2)".

  The arguments must be primitives of the same type, or tensors of the same
  dtype and similar shape, or sequences with the same sequence_dtype and
  sequence_shape.
  """

  def __init__(self):
    metadata = operation_base.OperationMetadata(
        docstring=TRIPLE_CREATION_DOCSTRING)
    super(TripleCreationOperation, self).__init__(
        num_args=3, weight=TRIPLE_CREATION_WEIGHT, metadata=metadata)

    for primitive_type in tf_coder_utils.PRIMITIVE_TYPES:
      self.add_value_filters(
          [filtering.get_type_filter(primitive_type)] * 3)

    def _tensor_value_filter(arg_value):
      """Only keeps values that are "small" tensors."""
      return (arg_value.is_tensor and
              arg_value.num_elements() * 3 <= limits.MAX_TENSOR_ELEMENTS)
    self.add_value_filters([_tensor_value_filter] * 3)

    self.add_value_filters([filtering.TENSOR_LIKE_SEQUENCE_FILTER] * 3)

    def _apply_filter(arg_values):
      """Ensures dtype and shape compatibility."""
      first, second, third = arg_values
      if first.is_tensor:  # Implies second and third are also tensors.
        if not first.dtype == second.dtype == third.dtype:
          return False
        shape_1 = first.shape
        shape_2 = second.shape
        shape_3 = third.shape
        if not len(shape_1) == len(shape_2) == len(shape_3):
          return False
        num_different = sum(
            not len_1 == len_2 == len_3
            for len_1, len_2, len_3 in zip(shape_1, shape_2, shape_3))
        return num_different <= 1
      elif first.is_sequence:  # Implies second and third is also sequences.
        return (first.sequence_dtype == second.sequence_dtype
                == third.sequence_dtype and
                first.sequence_shape == second.sequence_shape
                == third.sequence_shape)
      else:  # Implies that all args are same-type primitives.
        return True

    self.set_apply_filter(_apply_filter)

  def apply(self, arg_values, settings):
    """See base class."""
    try:
      return value.OperationValue(
          (arg_values[0].value, arg_values[1].value, arg_values[2].value),
          self, arg_values)
    except ValueError:
      return None

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return '(' + ', '.join(arg_strings) + ')'


@six.add_metaclass(abc.ABCMeta)
class SlicingBaseOperation(operation_base.Operation):
  """Common functionality for slicing operations."""

  def __init__(self):
    num_args = self._num_args()
    min_rank = self._min_rank()
    docstring = self._get_docstring()
    metadata = operation_base.OperationMetadata(docstring=docstring)
    super(SlicingBaseOperation, self).__init__(
        num_args=num_args, weight=SLICING_WEIGHT, metadata=metadata)
    self.add_value_filters(
        [filtering.get_tensor_min_rank_filter(min_rank)] +
        [filtering.INT_OR_INT_TENSOR_FILTER] * (num_args - 1))
    self.set_apply_filter(self._get_apply_filter())

  @abc.abstractmethod
  def _num_args(self):
    """Returns the number of arguments used by this operation."""

  @abc.abstractmethod
  def _min_rank(self):
    """Returns the minimum allowable tensor rank for the first argument."""

  @abc.abstractmethod
  def _get_apply_filter(self):
    """Returns an apply filter."""

  @abc.abstractmethod
  def _get_docstring(self):
    """Returns a docstring for this operation."""

  @abc.abstractmethod
  def _evaluate_slice(self, raw_values):
    """Evaluates the slice, e.g., `raw_values[0][raw_values[1] : ]`."""

  def apply(self, arg_values, settings):
    """See base class."""
    try:
      result = self._evaluate_slice(
          [arg_values[0].value] +
          [int(arg_value.value) for arg_value in arg_values[1:]])
      return value.OperationValue(result, self, arg_values)
    except Exception:  # pylint: disable=broad-except
      return None


class SlicingAxis0LeftOperation(SlicingBaseOperation):
  """An operation of the form "arg0[arg1:]"."""

  def _num_args(self):
    return 2

  def _min_rank(self):
    return 1

  def _get_apply_filter(self):
    def _check_index(arg_values):
      """Checks that the index is in range for the tensor."""
      tensor, index = arg_values
      length = tensor.shape[0]
      return -length <= int(index.value) < length
    return _check_index

  def _get_docstring(self):
    return SLICING_DOCSTRING_FORMAT.format(slice_kind='suffix', axis=0)

  def _evaluate_slice(self, raw_values):
    return raw_values[0][raw_values[1] : ]

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[' + arg_strings[1] + ':]'


class SlicingAxis0RightOperation(SlicingBaseOperation):
  """An operation of the form "arg0[:arg1]"."""

  def _num_args(self):
    return 2

  def _min_rank(self):
    return 1

  def _get_apply_filter(self):
    def _check_index(arg_values):
      """Checks that the index is in range for the tensor."""
      tensor, index = arg_values
      length = tensor.shape[0]
      return -length <= int(index.value) < length
    return _check_index

  def _get_docstring(self):
    return SLICING_DOCSTRING_FORMAT.format(slice_kind='prefix', axis=0)

  def _evaluate_slice(self, raw_values):
    return raw_values[0][ : raw_values[1]]

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[:' + arg_strings[1] + ']'


class SlicingAxis0BothOperation(SlicingBaseOperation):
  """An operation of the form "arg0[arg1:arg2]"."""

  def _num_args(self):
    return 3

  def _min_rank(self):
    return 1

  def _get_apply_filter(self):
    def _check_indices(arg_values):
      """Checks that the indices are in range for the tensor."""
      tensor, left_value, right_value = arg_values
      length = tensor.shape[0]
      left = int(left_value.value)
      right = int(right_value.value)
      if left < 0:
        left += length
      if right < 0:
        right += length
      return 0 <= left < right < length
    return _check_indices

  def _get_docstring(self):
    return SLICING_DOCSTRING_FORMAT.format(slice_kind='range', axis=0)

  def _evaluate_slice(self, raw_values):
    return raw_values[0][raw_values[1] : raw_values[2]]

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[' + arg_strings[1] + ':' + arg_strings[2] + ']'


class SlicingAxis1LeftOperation(SlicingBaseOperation):
  """An operation of the form "arg0[:, arg1:]"."""

  def _num_args(self):
    return 2

  def _min_rank(self):
    return 2

  def _get_apply_filter(self):
    def _check_index(arg_values):
      """Checks that the index is in range for the tensor."""
      tensor, index = arg_values
      length = tensor.shape[1]
      return -length <= int(index.value) < length
    return _check_index

  def _get_docstring(self):
    return SLICING_DOCSTRING_FORMAT.format(slice_kind='suffix', axis=1)

  def _evaluate_slice(self, raw_values):
    return raw_values[0][:, raw_values[1] : ]

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[:, ' + arg_strings[1] + ':]'


class SlicingAxis1RightOperation(SlicingBaseOperation):
  """An operation of the form "arg0[:, :arg1]"."""

  def _num_args(self):
    return 2

  def _min_rank(self):
    return 2

  def _get_apply_filter(self):
    def _check_index(arg_values):
      """Checks that the index is in range for the tensor."""
      tensor, index = arg_values
      length = tensor.shape[1]
      return -length <= int(index.value) < length
    return _check_index

  def _get_docstring(self):
    return SLICING_DOCSTRING_FORMAT.format(slice_kind='prefix', axis=1)

  def _evaluate_slice(self, raw_values):
    return raw_values[0][:, : raw_values[1]]

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[:, :' + arg_strings[1] + ']'


class SlicingAxis1BothOperation(SlicingBaseOperation):
  """An operation of the form "arg0[:, arg1:arg2]"."""

  def _num_args(self):
    return 3

  def _min_rank(self):
    return 2

  def _get_apply_filter(self):
    def _check_indices(arg_values):
      """Checks that the indices are in range for the tensor."""
      tensor, left_value, right_value = arg_values
      length = tensor.shape[1]
      left = int(left_value.value)
      right = int(right_value.value)
      if left < 0:
        left += length
      if right < 0:
        right += length
      return 0 <= left < right < length
    return _check_indices

  def _get_docstring(self):
    return SLICING_DOCSTRING_FORMAT.format(slice_kind='range', axis=1)

  def _evaluate_slice(self, raw_values):
    return raw_values[0][:, raw_values[1] : raw_values[2]]

  def reconstruct_expression_from_strings(self, arg_strings):
    """See base class."""
    return arg_strings[0] + '[:, ' + arg_strings[1] + ':' + arg_strings[2] + ']'

# TODO(kshi): Implement .shape attribute access. However, we already have a
# tf.shape() operation, so perhaps .shape access is not necessary.



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Defines the FilterGroup enum.

Every function used in TF-Coder is associated with one such FilterGroup.
"""

import enum


# LINT.IfChange(FilterGroup)
@enum.unique
class FilterGroup(enum.Enum):
  """A group of similar operations that should have the same filters.

  The number of arguments is crucially important when adding filters, so by
  convention the enum names have the number of arguments at the end.
  """
  # No filters. Even if some filtering might be reasonable, it could be faster
  # to just try all values to avoid the filtering overhead.
  NONE = 'NONE'

  #############################
  # Operations with 1 argument.

  # The argument is a shape.
  SHAPE_1 = 'SHAPE_1'
  # The argument is a sequence.
  SEQUENCE_1 = 'SEQUENCE_1'
  # The argument is a tensor.
  TENSOR_1 = 'TENSOR_1'
  # The argument is a float tensor.
  FLOATTENSOR_1 = 'FLOATTENSOR_1'
  # The argument is an int or float tensor.
  NUMERICTENSOR_1 = 'NUMERICTENSOR_1'
  # The argument is a boolean tensor.
  BOOLTENSOR_1 = 'BOOLTENSOR_1'
  # The argument is a SparseTensor.
  SPARSE_1 = 'SPARSE_1'
  # The argument is not a tensor.
  NOT_TENSOR_1 = 'NOT_TENSOR_1'
  # The argument is a 1-D tensor.
  TENSOR_1D_1 = 'TENSOR_1D_1'
  # The argument is a primitive or tensor.
  PRIMITIVE_OR_TENSOR_1 = 'PRIMITIVE_OR_TENSOR_1'

  ################################
  # Operations with 2 arguments.

  # The first argument is a tensor, and the second argument is an int
  # representing an axis, i.e., an int in the range [1, rank_of_tensor).
  TENSOR_AXIS_2 = 'TENSOR_AXIS_2'
  # The first argument is a tensor, and the second argument is an axis sequence.
  TENSOR_AXISSEQUENCE_2 = 'TENSOR_AXISSEQUENCE_2'
  # The first argument is a boolean tensor, the second is an axis.
  BOOLTENSOR_AXIS_2 = 'BOOLTENSOR_AXIS_2'
  # The first argument is an int or float tensor, the second is an axis.
  NUMERICTENSOR_AXIS_2 = 'NUMERICTENSOR_AXIS_2'
  # The first argument is a sequence of tensors, the second is an axis.
  TENSORSEQUENCE_AXIS_2 = 'TENSORSEQUENCE_AXIS_2'
  # The first argument is a tensor, the second is a shape.
  TENSOR_SHAPE_2 = 'TENSOR_SHAPE_2'
  # The first argument is a shape, the second is a primitive.
  SHAPE_PRIMITIVE_2 = 'SHAPE_PRIMITIVE_2'
  # The first argument is a tensor, the second is a boolean tensor.
  TENSOR_BOOLTENSOR_2 = 'TENSOR_BOOLTENSOR_2'
  # The second argument is a dtype. If the dtype is integral, the first argument
  # must be safely castable to that dtype.
  CASTABLE_DTYPE_2 = 'CASTABLE_DTYPE_2'
  # The two arguments are numeric (int or float) tensors with the same dtype,
  # and the two tensors are broadcastable, or one of the arguments is an int or
  # float primitive.
  SAME_DTYPE_NUMERIC_BROADCASTABLE_2 = 'SAME_DTYPE_NUMERIC_BROADCASTABLE_2'
  # The two arguments are float tensors with the same dtype and the two tensors
  # are broadcastable, or one of the arguments is a float primitive.
  SAME_DTYPE_FLOAT_BROADCASTABLE_2 = 'SAME_DTYPE_FLOAT_BROADCASTABLE_2'
  # The two arguments are boolean tensors with broadcastable shapes.
  BOOLTENSOR_BROADCASTABLE_2 = 'BOOLTENSOR_BROADCASTABLE_2'
  # Of the two arguments, one must be a SparseTensor, and the other must be a
  # Tensor or SparseTensor. The two arguments must have the same shape.
  SAME_SHAPE_ONE_SPARSE_2 = 'SAME_SHAPE_ONE_SPARSE_2'
  # The two arguments arguments must be SparseTensors with the same shape.
  SAME_SHAPE_BOTH_SPARSE_2 = 'SAME_SHAPE_BOTH_SPARSE_2'
  # The first argument is an int representing an axis, the second is a list of
  # SparseTensor.
  AXIS_SPARSESEQUENCE_2 = 'AXIS_SPARSESEQUENCE_2'
  # The first argument is a SparseTensor, the second is an axis.
  SPARSE_AXIS_2 = 'SPARSE_AXIS_2'
  # The first argument is a SparseTensor, the second is a shape.
  SPARSE_SHAPE_2 = 'SPARSE_SHAPE_2'
  # The first argument is a SparseTensor, the second is a primitive.
  SPARSE_PRIMITIVE_2 = 'SPARSE_PRIMITIVE_2'
  # The first argument is an int or float tensor, and the second argument is a
  # 1D int tensor with nonnegative nondecreasing entries, and the two arguments
  # have equal lengths in the first dimension.
  SEGMENT_OPERATION_2 = 'SEGMENT_OPERATION_2'

  ##############################
  # Operations with 3 arguments.

  # The first argument is a SparseTensor, the second is an int, and the third is
  # an axis.
  SPARSE_INT_AXIS_3 = 'SPARSE_INT_AXIS_3'
  # The first argument is a SparseTensor, the second is an axis, the third is a
  # boolean.
  SPARSE_AXIS_BOOL_3 = 'SPARSE_AXIS_BOOL_3'
  # The first argument is an int or float tensor, the second argument is an int
  # tensor with nonnegative entries and with a shape that is a prefix of the
  # first argument's shape, and the third argument is an integer greater than
  # the maximum element of the second argument.
  UNSORTED_SEGMENT_OPERATION_3 = 'UNSORTED_SEGMENT_OPERATION_3'

  #########################################
  # Operations with other special handling.

  # The argument contains nonnegative ints with a small maximum.
  BINCOUNT_1 = 'BINCOUNT_1'
  # The argument results in a small tensor.
  EYE_1 = 'EYE_1'
  # The argument results in a small tensor.
  RANGE_1 = 'RANGE_1'
  # The argument contains a few ints with a small maximum.
  SEQUENCE_MASK_1 = 'SEQUENCE_MASK_1'
  # The argument is a tensor with at least one squeezable dimension.
  SQUEEZE_1 = 'SQUEEZE_1'

  # The first argument is a tensor, the second is a compatible shape.
  BROADCAST_TO_2 = 'BROADCAST_TO_2'
  # The first argument is a tensor, the second is an axis in the range
  # [-1, rank_of_tensor]. Note that this range is slightly different from the
  # TENSOR_AXIS_2 filter.
  EXPAND_DIMS_2 = 'EXPAND_DIMS_2'
  # The arguments result in a small tensor.
  EYE_ROWS_COLS_2 = 'EYE_ROWS_COLS_2'
  # The arguments result in a small tensor.
  EYE_ROWS_DTYPE_2 = 'EYE_ROWS_DTYPE_2'
  # The first argument is a tensor, the second is a tensor containing ints
  # suitable for indexing into the first tensor on axis 0.
  GATHER_2 = 'GATHER_2'
  # The first argument is a tensor, the second is a tensor containing ints
  # suitable for indexing into the first tensor on multiple dimensions.
  GATHER_ND_2 = 'GATHER_ND_2'
  # Ensures the tensors are both numeric and have the same dtype and rank.
  MATMUL_2 = 'MATMUL_2'
  # Ensures that tf.one_hot(indices, depth) produces a small result.
  ONE_HOT_2 = 'ONE_HOT_2'
  # The first argument must be a tensor, and the second must be a nested int
  # list or int32 tensor of shape [rank_of_arg_1, 2].
  PAD_2 = 'PAD_2'
  # The first argument is a tensor or scalar, and the second argument is a
  # nonnegative 1D int tensor/sequence (or nonnegative int scalar) with either
  # length 1, or length equal to the number of elements in the first argument.
  REPEAT_2 = 'REPEAT_2'
  # The first argument is sorted in the last dimension, the second argument is
  # the same dtype and rank, and all dimension lengths match except the last.
  SEARCHSORTED_2 = 'SEARCHSORTED_2'
  # The first argument contains a few ints with a small maximum, and the second
  # is an int primitive or tensor with a value in range.
  SEQUENCE_MASK_2 = 'SEQUENCE_MASK_2'
  # Ensures that tf.sparse.retain(sp_input, to_retain) has arguments that match,
  # i.e., to_retain is 1D, contains booleans, and has the right length.
  SPARSE_RETAIN_2 = 'SPARSE_RETAIN_2'
  # Ensures that tf.sparse.to_indicator(sp_input, vocab_size) produces a small
  # result.
  SPARSE_TO_INDICATOR_2 = 'SPARSE_TO_INDICATOR_2'
  # The first argument is a SparseTensor, the second is 1D (ints) with correct
  # length.
  SPARSE_TRANSPOSE_2 = 'SPARSE_TRANSPOSE_2'
  # The first argument is a tensor with more than 1 squeezable dimension, and
  # the second argument is an int specifying a squeezable dimension.
  SQUEEZE_2 = 'SQUEEZE_2'
  # Ensures that tf.tile(input, multiples) produces a small result.
  TILE_2 = 'TILE_2'
  # The first argument is a numeric tensor, the second is a compatible int
  # primitive or tensor.
  TOP_K_2 = 'TOP_K_2'
  # The first argument is a tensor, the second is 1D (ints) with correct length.
  TRANSPOSE_2 = 'TRANSPOSE_2'

  # The first argument must be an int or float tensor, and the second and third
  # arguments must both be nondecreasing primitives.
  CLIP_BY_VALUE_3 = 'CLIP_BY_VALUE_3'
  # The first argument is a tensor, the second is a tensor containing ints
  # suitable for indexing into the first tensor on multiple dimensions, and the
  # third is a number of batch dimensions.
  GATHER_ND_3 = 'GATHER_ND_3'
  # The first argument must be a tensor, the second must be a nested int list or
  # int32 tensor of shape [rank_of_arg_1, 2], and the third must be a scalar of
  # the same type as the first argument.
  PAD_3 = 'PAD_3'
  # The arguments result in a small tensor.
  RANGE_3 = 'RANGE_3'
  # The first argument is a nonscalar tensor, the third argument is an axis in
  # range, and the second argument is a nonnegative 1D int tensor/sequence (or
  # nonnegative int scalar) with either length 1 or length equal to the length
  # of the first argument at the specified axis.
  REPEAT_3 = 'REPEAT_3'
  # The second and third arguments must be int primitives, lists of ints, or 1D
  # int tensors, and they must have the same shape.
  ROLL_3 = 'ROLL_3'
  # In tf.scatter_nd(indices, updates, shape), `indices` must be an int tensor
  # representing valid indices, `shape` must be a shape tuple, and `updates`
  # must have a compatible shape with `indices` and `shape`.
  SCATTER_ND_3 = 'SCATTER_ND_3'
  # In tf.sparse.slice(sp_input, start, size), the arguments start and size must
  # be 1D (ints) with the same length as sp_input.dense_shape.
  SPARSE_SLICE_3 = 'SPARSE_SLICE_3'
  # The arguments must have the right shape to make a SparseTensor.
  SPARSETENSOR_3 = 'SPARSETENSOR_3'
  # In tf.tensor_scatter_nd_update(tensor, indices, updates), `indices` must be
  # an int tensor representing valid indices, and `updates` must have a
  # compatible shape and be the same dtype as `tensor`.
  TENSOR_SCATTER_ND_UPDATE_3 = 'TENSOR_SCATTER_ND_UPDATE_3'
  # The first two arguments are tensors with the same dtype, and the third
  # contains ints of the appropriate shape.
  TENSORDOT_3 = 'TENSORDOT_3'
  # Ensures that the shapes and dtypes for tf.where(condition, x, y) match.
  WHERE_3 = 'WHERE_3'

  # Ensures that params, indices, axis, and batch_dims are all compatible
  # according to the tf.gather documentation.
  GATHER_4 = 'GATHER_4'

# LINT.ThenChange(value_search/operation_filtering.py:add_filters_to_function_operation)



# Copyright 2021 The TF-Coder Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Tests for all_benchmarks.py."""

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
import tensorflow as tf
from tf_coder import tf_coder_utils
from tf_coder.benchmarks import all_benchmarks
from tf_coder.benchmarks import google_benchmarks
from tf_coder.benchmarks import stackoverflow_benchmarks
from tf_coder.benchmarks import test_benchmarks
from tf_coder.value_search import value as value_module
from tf_coder.value_search import value_search


def _parameterize_all_benchmarks():
  """Creates parameterized test cases for all benchmarks.

  This is useful so that tests can run on each benchmark individually, instead
  of having a single test loop over all benchmarks. In this way, issues with
  multiple benchmarks can be identified in one round of testing, and it is
  clearer which of the benchmarks need further attention.

  Returns:
    A list of tuples (test_case_name, benchmark, use_eager) for all benchmarks
    and both values of use_eager (True and False).
  """
  parameterized_tuples = []
  for index, benchmark in enumerate(all_benchmarks.all_benchmarks()):
    # The index ensures all test cases have distinct names, even if multiple
    # benchmarks have the same name.
    test_case_name = '{index}_{name}'.format(index=index, name=benchmark.name)
    parameterized_tuples.append((test_case_name, benchmark))
  return parameterized_tuples


class AllBenchmarksTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('including_ignored', True, ['test_add', 'test_cast',
                                   'inconsistent_target_program', 'test_add']),
      ('not_including_ignored', False, ['test_add', 'test_cast']))
  def test_all_benchmarks_finds_correct_functions(self, include_ignored,
                                                  expected_names):
    benchmarks = all_benchmarks.all_benchmarks(modules=[test_benchmarks],
                                               include_ignored=include_ignored)
    benchmark_names = [benchmark.name for benchmark in benchmarks]
    self.assertCountEqual(benchmark_names, expected_names)

  def test_all_benchmarks_have_unique_names(self):
    names = []
    for benchmark in all_benchmarks.all_benchmarks(include_ignored=False):
      name = benchmark.name
      self.assertIsNotNone(name)
      names.append(name)
    self.assertLen(set(names), len(names))

  def test_all_benchmarks_have_source(self):
    for benchmark in all_benchmarks.all_benchmarks(include_ignored=False):
      self.assertIsNotNone(benchmark.source)

  def test_all_benchmarks_have_description(self):
    modules = [google_benchmarks, stackoverflow_benchmarks]
    for benchmark in all_benchmarks.all_benchmarks(modules=modules):
      self.assertTrue(benchmark.description)

  def test_all_benchmarks_have_reason_if_ignored(self):
    for benchmark in all_benchmarks.all_benchmarks(include_ignored=True):
      if benchmark.should_ignore:
        self.assertIsNotNone(benchmark.ignore_reason)

  @parameterized.named_parameters(
      ('zero_occurrences', 'nonexistent_name', True),
      ('one_occurrence', 'test_cast', False),
      ('two_occurrences', 'test_add', True))
  def test_find_benchmark_with_name(self, benchmark_name, expect_none):
    result = all_benchmarks.find_benchmark_with_name(
        benchmark_name, include_ignored=True, modules=[test_benchmarks])
    if expect_none:
      self.assertIsNone(result)
    else:
      self.assertIsNotNone(result)
      self.assertEqual(result.name, benchmark_name)

  def test_get_chosen_benchmarks_single_success(self):
    result = all_benchmarks.get_chosen_benchmarks(
        'test_add', include_ignored=False, modules=[test_benchmarks])
    self.assertLen(result, 1)
    self.assertEqual(result[0].name, 'test_add')

  def test_get_chosen_benchmarks_all_success(self):
    result = all_benchmarks.get_chosen_benchmarks(
        'ALL', modules=[test_benchmarks])
    self.assertGreater(len(result), 1)

  def test_get_chosen_benchmarks_duplicate_name(self):
    result = all_benchmarks.get_chosen_benchmarks(
        'test_add', include_ignored=True, modules=[test_benchmarks])
    self.assertEmpty(result)


class StackOverflowBenchmarksTest(absltest.TestCase):

  def test_all_stackoverflow_benchmarks_have_unique_source(self):
    all_sources = set()
    for benchmark in all_benchmarks.all_benchmarks(
        modules=[stackoverflow_benchmarks], include_ignored=False):
      self.assertNotIn(benchmark.source, all_sources)
      all_sources.add(benchmark.source)


class TargetProgramTest(parameterized.TestCase):

  def _check_target_program(self, benchmark):
    """Checks that a benchmark's target program is consistent with its examples.

    Args:
      benchmark: A Benchmark to verify.
    """
    self.assertIsNotNone(benchmark.target_program)

    for example in benchmark.examples:
      # Turn inputs into constant tensors and assign them to variables using a
      # new global namespace.
      global_namespace = {'tf': tf}
      input_names_to_objects = value_search._input_names_to_objects(
          example.inputs)
      for input_name, input_object in input_names_to_objects.items():
        input_value = value_module.InputValue(input_object, name='dummy_name')
        global_namespace[input_name] = input_value.value

      # Evaluate the target program, which uses the canonical variables.
      target_program_output = eval(benchmark.target_program, global_namespace)  # pylint: disable=eval-used

      # Check that the two outputs have equal string representation.
      expected_output = tf_coder_utils.convert_to_tensor(example.output)
      self.assertEqual(
          tf_coder_utils.object_to_string(expected_output),
          tf_coder_utils.object_to_string(target_program_output))

  def test_check_target_program_passes_for_good_benchmarks(self):
    good_benchmarks = [test_benchmarks.test_add(), test_benchmarks.test_cast()]
    for good_benchmark in good_benchmarks:
      self._check_target_program(good_benchmark)

  def test_check_target_program_fails_for_bad_benchmarks(self):
    bad_benchmark = test_benchmarks.inconsistent_target_program()
    with self.assertRaises(AssertionError):
      self._check_target_program(bad_benchmark)

  @parameterized.named_parameters(_parameterize_all_benchmarks())
  def test_all_benchmarks_have_correct_target_programs(self, benchmark):
    # Some benchmarks don't have target programs because we don't know how to
    # solve them ourselves.
    if benchmark.target_program is not None:
      self._check_target_program(benchmark)


if __name__ == '__main__':
  logging.set_verbosity(logging.ERROR)

  absltest.main()
